import numpy
import os.path as osp

from six import BytesIO

from smqtk.representation import DescriptorElement
from smqtk.utils import file_utils
from smqtk.utils.string_utils import partition_string


class DescriptorMemoryElement (DescriptorElement):
    """
    In-memory representation of descriptor elements. Stored vectors are
    effectively immutable.

    Example
    -------
    >>> self = DescriptorMemoryElement('random', 0)
    """

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self, type_str, uuid):
        super(DescriptorMemoryElement, self).__init__(type_str, uuid)
        self.__v = None

    def __getstate__(self):
        state = super(DescriptorMemoryElement, self).__getstate__()
        # save vector as binary string
        b = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(b, self.vector())
        state['v'] = b.getvalue()
        return state

    def __setstate__(self, state):
        # Handle the previous state format:
        if isinstance(state, tuple):
            self._type_label = state[0]
            self._uuid = state[1]
            b = BytesIO(state[2])
        else:  # dictionary
            super(DescriptorMemoryElement, self).__setstate__(state)
            b = BytesIO(state['v'])
        self.__v = numpy.load(b)

    def get_config(self):
        """
        :return: JSON type compliant configuration dictionary.
        :rtype: dict
        """
        return {}

    def has_vector(self):
        """
        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool
        """
        return self.__v is not None

    def vector(self):
        """
        Implementation Note
        -------------------
        A copy of the internally stored vector is returned.

        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.core.multiarray.ndarray or None

        """
        # Copy if storing an array, otherwise return the None value
        if self.__v is not None:
            return numpy.copy(self.__v)
        return None

    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        ``new_vec`` may be None, which clears this descriptor's vector from the
        cache.

        Implementation Note
        -------------------
        This implementation copies input arrays before storage to mimic
        immutability.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.core.multiarray.ndarray | tuple | list | None

        :returns: Self.
        :rtype: DescriptorMemoryElement

        """
        # Copy a non-None value given, otherwise stay None
        if new_vec is not None:
            self.__v = numpy.copy(new_vec)
        else:
            self.__v = None
        return self


class DescriptorFileElement (DescriptorElement):
    """
    File-based storage of descriptor element.

    When initialized, saves uuid and vector as a serialized pickle-file and
    numpy-format npy file, respectively. These are named according to the string
    representation of the uuid object provided. These are then loaded from disk
    when the ``uuid`` or ``vector`` methods are called. This is in turn slower
    performance wise than ``MemoryElement``, however RAM consumption will be
    lower for large number of elements that would otherwise exceed RAM storage
    space.

    """

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self, type_str, uuid, save_dir, subdir_split=None):
        """
        Initialize a file-base descriptor element.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: uuid for this descriptor
        :type uuid: collections.Hashable

        :param save_dir: Directory to save this element's contents. If this path
            is relative, we interpret as relative to the current working
            directory.
        :type save_dir: str | unicode

        :param subdir_split: If a positive integer and greater than 1, this will
            cause us to store the vector file in a subdirectory under the
            ``save_dir`` based on our ``uuid``. The integer value specifies the
            number of splits that we will make in the stringification of this
            descriptor's UUID. The last split component is left off when
            determining the save directory (thus the >1 above).

            Dashes are stripped from this string (as would happen if given an
            uuid.UUID instance as the uuid element).
        :type subdir_split: None | int

        """
        super(DescriptorFileElement, self).__init__(type_str, uuid)
        self._save_dir = osp.abspath(osp.expanduser(save_dir))
        self._subdir_split = subdir_split

        # Generate filepath from parameters
        if self._subdir_split and int(self._subdir_split) > 1:
            save_dir = osp.join(
                self._save_dir,
                *partition_string(str(self.uuid()).replace('-', ''),
                                  int(self._subdir_split))[:-1]
            )
        else:
            save_dir = self._save_dir
        self._vec_filepath = osp.join(save_dir,
                                      "%s.%s.vector.npy" % (self.type(),
                                                            str(self.uuid())))

    def __getstate__(self):
        state = super(DescriptorFileElement, self).__getstate__()
        state.update({
            '_save_dir': self._save_dir,
            '_subdir_split': self._subdir_split,
            '_vec_filepath': self._vec_filepath,
        })
        return state

    def __setstate__(self, state):
        super(DescriptorFileElement, self).__setstate__(state)
        self._save_dir = state['_save_dir']
        self._subdir_split = state['_subdir_split']
        self._vec_filepath = state['_vec_filepath']

    def get_config(self):
        return {
            "save_dir": self._save_dir,
            'subdir_split': self._subdir_split
        }

    def has_vector(self):
        """
        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool
        """
        return osp.isfile(self._vec_filepath)

    def vector(self):
        """
        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.core.multiarray.ndarray or None
        """
        # TODO: Load as memmap?
        #       i.e. modifications by user to vector will be reflected on disk.
        if self.has_vector():
            return numpy.load(self._vec_filepath)
        else:
            return None

    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.core.multiarray.ndarray

        :returns: Self.
        :rtype: DescriptorFileElement

        """
        file_utils.safe_create_dir(osp.dirname(self._vec_filepath))
        numpy.save(self._vec_filepath, new_vec)
        return self


DESCRIPTOR_ELEMENT_CLASS = [
    DescriptorMemoryElement,
    DescriptorFileElement,
]



from __future__ import division, print_function
from math import sqrt
import unittest

import numpy
import six
from six import BytesIO

from smqtk.algorithms.nn_index.lsh.functors.itq import ItqFunctor
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.utils import merge_dict


class TestItqFunctor (unittest.TestCase):

    def test_is_usable(self):
        # Should always be usable due to no non-standard dependencies.
        self.assertTrue(ItqFunctor.is_usable())

    def test_default_configuration(self):
        c = ItqFunctor.get_default_config()
        self.assertEqual(ItqFunctor.from_config(c).get_config(), c)

    def test_configuration_with_caches(self):
        expected_mean_vec = numpy.array([1, 2, 3])
        expected_rotation = numpy.eye(3)

        expected_mean_vec_bytes = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(expected_mean_vec_bytes, expected_mean_vec)
        expected_mean_vec_bytes = expected_mean_vec_bytes.getvalue()

        expected_rotation_bytes = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(expected_rotation_bytes, expected_rotation)
        expected_rotation_bytes = expected_rotation_bytes.getvalue()

        new_parts = {
            'mean_vec_cache': {
                'DataMemoryElement': {
                    'bytes': expected_mean_vec_bytes
                },
                'type': 'DataMemoryElement'
            },
            'rotation_cache': {
                'DataMemoryElement': {
                    'bytes': expected_rotation_bytes
                },
                'type': 'DataMemoryElement'
            },
            'bit_length': 153,
            'itq_iterations': 7,
            'normalize': 2,
            'random_seed': 58,
        }
        c = merge_dict(ItqFunctor.get_default_config(), new_parts)

        itq = ItqFunctor.from_config(c)

        # Checking that loaded parameters were correctly set and cache elements
        # correctly return intended vector/matrix.
        numpy.testing.assert_equal(itq.mean_vec, [1, 2, 3])
        numpy.testing.assert_equal(itq.rotation, [[1, 0, 0],
                                                  [0, 1, 0],
                                                  [0, 0, 1]])
        self.assertEqual(itq.bit_length, 153)
        self.assertEqual(itq.itq_iterations, 7)
        self.assertEqual(itq.normalize, 2)
        self.assertEqual(itq.random_seed, 58)

    def test_norm_vector_no_normalization(self):
        itq = ItqFunctor(normalize=None)

        v = numpy.array([0, 1])
        numpy.testing.assert_array_equal(itq._norm_vector(v), v)

        v = numpy.array([[0, 1, 1, .4, .1]])
        numpy.testing.assert_array_equal(itq._norm_vector(v), v)

        v = numpy.array([0]*128)
        numpy.testing.assert_array_equal(itq._norm_vector(v), v)

    def test_norm_vector_n2(self):
        itq = ItqFunctor(normalize=2)

        v = numpy.array([1, 0])
        numpy.testing.assert_array_almost_equal(
            itq._norm_vector(v), [1, 0]
        )

        v = numpy.array([1, 1])
        numpy.testing.assert_array_almost_equal(
            itq._norm_vector(v), [1./sqrt(2), 1./sqrt(2)]
        )

    def test_get_config_no_cache(self):
        itq = ItqFunctor(bit_length=1, itq_iterations=2, normalize=3,
                         random_seed=4)
        c = itq.get_config()
        self.assertEqual(c['bit_length'], 1)
        self.assertEqual(c['itq_iterations'], 2)
        self.assertEqual(c['normalize'], 3)
        self.assertEqual(c['random_seed'], 4)
        self.assertIsNone(c['mean_vec_cache']['type'])
        self.assertIsNone(c['rotation_cache']['type'])

    def test_get_config_with_cache_elements(self):
        itq = ItqFunctor(bit_length=5, itq_iterations=6, normalize=7,
                         random_seed=8)
        itq.mean_vec_cache_elem = DataMemoryElement('cached vec bytes')
        itq.rotation_cache_elem = DataMemoryElement('cached rot bytes')

        c = itq.get_config()
        self.assertEqual(c['bit_length'], 5)
        self.assertEqual(c['itq_iterations'], 6)
        self.assertEqual(c['normalize'], 7)
        self.assertEqual(c['random_seed'], 8)
        self.assertEqual(c['mean_vec_cache']['type'], "DataMemoryElement")
        self.assertEqual(c['mean_vec_cache']['DataMemoryElement']['bytes'],
                         'cached vec bytes')
        self.assertEqual(c['rotation_cache']['DataMemoryElement']['bytes'],
                         'cached rot bytes')

    def test_has_model(self):
        itq = ItqFunctor()
        # with no vector/rotation set, should return false.
        self.assertFalse(itq.has_model())
        # If only one of the two is None, then false should be returned.
        itq.mean_vec = 'mean vec'
        itq.rotation = None
        self.assertFalse(itq.has_model())
        itq.mean_vec = None
        itq.rotation = 'rotation'
        self.assertFalse(itq.has_model())
        # If both are not None, return true.
        itq.mean_vec = 'mean vec'
        itq.rotation = 'rotation'
        self.assertTrue(itq.has_model())

    def test_save_model_no_caches(self):
        expected_mean_vec = numpy.array([1, 2, 3])
        expected_rotation = numpy.eye(3)

        # Cache variables should remain None after save.
        itq = ItqFunctor()
        itq.mean_vec = expected_mean_vec
        itq.rotation = expected_rotation
        itq.save_model()
        self.assertIsNone(itq.mean_vec_cache_elem)
        self.assertIsNone(itq.mean_vec_cache_elem)

    def test_save_model_with_read_only_cache(self):
        # If one or both cache elements are read-only, no saving.
        expected_mean_vec = numpy.array([1, 2, 3])
        expected_rotation = numpy.eye(3)

        itq = ItqFunctor()
        itq.mean_vec = expected_mean_vec
        itq.rotation = expected_rotation

        # read-only mean-vec cache
        itq.mean_vec_cache_elem = DataMemoryElement(readonly=True)
        itq.rotation_cache_elem = DataMemoryElement(readonly=False)
        itq.save_model()
        self.assertEqual(itq.mean_vec_cache_elem.get_bytes(), six.b(''))
        self.assertEqual(itq.rotation_cache_elem.get_bytes(), six.b(''))

        # read-only rotation cache
        itq.mean_vec_cache_elem = DataMemoryElement(readonly=False)
        itq.rotation_cache_elem = DataMemoryElement(readonly=True)
        itq.save_model()
        self.assertEqual(itq.mean_vec_cache_elem.get_bytes(), six.b(''))
        self.assertEqual(itq.rotation_cache_elem.get_bytes(), six.b(''))

        # Both read-only
        itq.mean_vec_cache_elem = DataMemoryElement(readonly=True)
        itq.rotation_cache_elem = DataMemoryElement(readonly=True)
        itq.save_model()
        self.assertEqual(itq.mean_vec_cache_elem.get_bytes(), six.b(''))
        self.assertEqual(itq.rotation_cache_elem.get_bytes(), six.b(''))

    def test_save_model_with_writable_caches(self):
        # If one or both cache elements are read-only, no saving.
        expected_mean_vec = numpy.array([1, 2, 3])
        expected_rotation = numpy.eye(3)

        expected_mean_vec_bytes = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(expected_mean_vec_bytes, expected_mean_vec)
        expected_mean_vec_bytes = expected_mean_vec_bytes.getvalue()

        expected_rotation_bytes = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(expected_rotation_bytes, expected_rotation)
        expected_rotation_bytes = expected_rotation_bytes.getvalue()

        itq = ItqFunctor()
        itq.mean_vec = expected_mean_vec
        itq.rotation = expected_rotation
        itq.mean_vec_cache_elem = DataMemoryElement(readonly=False)
        itq.rotation_cache_elem = DataMemoryElement(readonly=False)

        itq.save_model()
        self.assertEqual(itq.mean_vec_cache_elem.get_bytes(),
                         expected_mean_vec_bytes)
        self.assertEqual(itq.rotation_cache_elem.get_bytes(),
                         expected_rotation_bytes)

    def test_fit_has_model(self):
        # When trying to run fit where there is already a mean vector and
        # rotation set.
        itq = ItqFunctor()
        itq.mean_vec = 'sim vec'
        itq.rotation = 'sim rot'
        self.assertRaisesRegexp(
            RuntimeError,
            "Model components have already been loaded.",
            itq.fit, []
        )

    def test_fit_short_descriptors_for_bit_length(self):
        # Should error when input descriptors have fewer dimensions than set bit
        # length for output hash codes (limitation of PCA method currently
        # used).
        fit_descriptors = []
        for i in range(3):
            d = DescriptorMemoryElement(six.b('test'), i)
            d.set_vector([-1+i, -1+i])
            fit_descriptors.append(d)

        itq = ItqFunctor(bit_length=8)
        self.assertRaisesRegexp(
            ValueError,
            "Input descriptors have fewer features than requested bit encoding",
            itq.fit, fit_descriptors
        )
        self.assertIsNone(itq.mean_vec)
        self.assertIsNone(itq.rotation)

        # Should behave the same when input is an iterable
        self.assertRaisesRegexp(
            ValueError,
            "Input descriptors have fewer features than requested bit encoding",
            itq.fit, iter(fit_descriptors)
        )
        self.assertIsNone(itq.mean_vec)
        self.assertIsNone(itq.rotation)

    def test_fit(self):
        fit_descriptors = []
        for i in range(5):
            d = DescriptorMemoryElement(six.b('test'), i)
            d.set_vector([-2. + i, -2. + i])
            fit_descriptors.append(d)

        itq = ItqFunctor(bit_length=1, random_seed=0)
        itq.fit(fit_descriptors)

        # TODO: Explanation as to why this is the expected result.
        numpy.testing.assert_array_almost_equal(itq.mean_vec, [0, 0])
        numpy.testing.assert_array_almost_equal(itq.rotation, [[1 / sqrt(2)],
                                                               [1 / sqrt(2)]])
        self.assertIsNone(itq.mean_vec_cache_elem)
        self.assertIsNone(itq.rotation_cache_elem)

    def test_fit_with_cache(self):
        fit_descriptors = []
        for i in range(5):
            d = DescriptorMemoryElement(six.b('test'), i)
            d.set_vector([-2. + i, -2. + i])
            fit_descriptors.append(d)

        itq = ItqFunctor(DataMemoryElement(), DataMemoryElement(),
                         bit_length=1, random_seed=0)
        itq.fit(fit_descriptors)

        # TODO: Explanation as to why this is the expected result.
        numpy.testing.assert_array_almost_equal(itq.mean_vec, [0, 0])
        numpy.testing.assert_array_almost_equal(itq.rotation, [[1 / sqrt(2)],
                                                               [1 / sqrt(2)]])
        self.assertIsNotNone(itq.mean_vec_cache_elem)
        numpy.testing.assert_array_almost_equal(
            numpy.load(BytesIO(itq.mean_vec_cache_elem.get_bytes())),
            [0, 0]
        )

        self.assertIsNotNone(itq.rotation_cache_elem)
        numpy.testing.assert_array_almost_equal(
            numpy.load(BytesIO(itq.rotation_cache_elem.get_bytes())),
            [[1 / sqrt(2)],
             [1 / sqrt(2)]]
        )

    def test_get_hash(self):
        fit_descriptors = []
        for i in range(5):
            d = DescriptorMemoryElement(six.b('test'), i)
            d.set_vector([-2. + i, -2. + i])
            fit_descriptors.append(d)

        # The following "rotation" matrix should cause any 2-feature descriptor
        # to the right of the line ``y = -x`` to be True, and to the left as
        # False. If on the line, should be True.
        itq = ItqFunctor(bit_length=1, random_seed=0)
        itq.mean_vec = numpy.array([0., 0.])
        itq.rotation = numpy.array([[1. / sqrt(2)],
                                    [1. / sqrt(2)]])

        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([1, 1])), [True])
        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([-1, -1])), [False])

        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([-1, 1])), [True])
        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([-1.001, 1])), [False])
        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([-1, 1.001])), [True])

        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([1, -1])), [True])
        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([1, -1.001])), [False])
        numpy.testing.assert_array_equal(
            itq.get_hash(numpy.array([1.001, -1])), [True])



import io
import logging
import PIL.Image

from smqtk.representation.data_element.file_element import DataElement


def is_loadable_image(data_element):
    """
    Determine if an image is able to be loaded by PIL.

    :param data_element: A data element to check
    :type data_element: DataElement

    :return: Whether or not the image is loadable
    :rtype: bool

    Example:
    >>>

    """
    log = logging.getLogger(__name__)

    try:
        PIL.Image.open(io.BytesIO(data_element.get_bytes()))
        return True
    except IOError as ex:
        # noinspection PyProtectedMember
        log.debug("Failed to convert '%s' bytes into an image "
                  "(error: %s). Skipping", data_element, str(ex))
        return False


def is_valid_element(data_element, valid_content_types=None, check_image=False):
    """
    Determines if a given data element is valid.

    :param data_element: Data element
    :type data_element: DataElement

    :param valid_content_types: List of valid content types, or None to skip
        content type checking.
    :type valid_content_types: iterable | None

    :param check_image: Whether or not to try loading the image with PIL. This
        often catches issues that content type can't, such as corrupt images.
    :type check_image: bool

    :return: Whether or not the data element is valid
    :rtype: bool

    """
    log = logging.getLogger(__name__)

    if (valid_content_types is not None and
            data_element.content_type() not in valid_content_types):
        log.debug("Skipping file (invalid content) type for "
                  "descriptor generator (data_element='%s', ct=%s)",
                  data_element, data_element.content_type())
        return False

    if check_image and not is_loadable_image(data_element):
        return False

    return isinstance(data_element, DataElement)




class NoClassificationError (Exception):
    """
    When a ClassificationElement has no mapping yet set, but an operation
    required it.
    """


class ReadOnlyError (Exception):
    """
    For when an attempt at modifying an immutable container is made.
    """


class NoUriResolutionError (Exception):
    """
    Standard exception thrown by base DataElement from_uri method when a
    subclass does not implement URI resolution.
    """


class InvalidUriError (Exception):
    """
    An invalid URI was provided.
    """

    def __init__(self, uri_value, reason):
        super(InvalidUriError, self).__init__(uri_value, reason)
        self.uri = uri_value
        self.reason = reason


class MissingLabelError(Exception):
    """
    Raised by ClassifierCollection.classify when requested classifier labels
    are missing from collection.
    """
    def __init__(self, labels):
        """
        :param labels: The labels missing from the collection
        :type labels: set[str]
        """
        super(MissingLabelError, self).__init__(labels)
        self.labels = labels



"""
LICENCE
-------
Copyright 2013-2015 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.
"""

from collections import deque
import multiprocessing
from threading import current_thread
import time


class ReaderUpdateException (Exception):
    """
    Exception thrown when an acquireWrite() call is called within a reader lock
    while another reader lock is already upgraded.
    """
    pass


class DummyRWLock (object):
    """
    Dummy object that mimics the API of a ReadWrite lock but doesn't actually
    do anything.
    """

    def acquireRead(self, timeout=None, _id=None):
        pass

    def acquireWrite(self, timeout=None, _id=None):
        pass

    def releaseRead(self, _id=None):
        pass

    def releaseWrite(self, _id=None):
        pass

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def read_lock(self, timeout=None, _id=None):
        # noinspection PyMethodParameters
        class DummyReadWithLock (object):
            def __enter__(_self):
                pass

            def __exit__(_self, *args):
                pass
        return DummyReadWithLock()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def write_lock(self, timeout=None, _id=None):
        # noinspection PyMethodParameters
        class DummyWriteWithLock (object):
            def __enter__(_self):
                pass

            def __exit__(_self, *args):
                pass
        return DummyWriteWithLock()


class ReadWriteLock (object):
    """ Reentrant Read-Write for multiprocessing

    Allows multiple threads/processes to simultaneously hold a read lock, while
    allowing only a single thread/process to hold a write lock at the same point
    of time.

    WARNING: This lock is NOT write-lock re-entrant. Attempting to acquire a
    write lock within a write lock WILL cause a dead-lock.

    When a read lock is requested while a write lock is held, the reader
    is blocked; when a write lock is requested while another write lock is
    held or there are read locks, the writer is blocked.

    When resolving who acquires the lock between readers and writers, writers
    are always preferred by this implementation: if there are blocked
    threads waiting for a write lock, current readers may request more read
    locks (which they eventually should free, as they starve the waiting
    writers otherwise), but a new thread requesting a read lock will not
    be granted one, and block. This might mean starvation for readers if
    two writer threads interweave their calls to acquireWrite() without
    leaving a window only for readers.

    NOT IMPLEMENTED :: TODO ::
    In case a current reader requests a write lock, this can and will be
    satisfied without giving up the read locks first, but, only one thread
    may perform this kind of lock upgrade, as a deadlock would otherwise
    occur. After the write lock has been granted, the thread will hold a
    full write lock, and not be downgraded after the upgrading call to
    acquireWrite() has been match by a corresponding release().
        - could add a acquireWriteUpgrade method for the specific purpose of
          acquiring a write lock in the presence of readers.

    """

    def __init__(self):
        """ Initialize lock state
        """
        # NOTE: If ever we want an equal access priority instead of giving
        # access priority to the writers, we could have a global queue FIFO
        # queue instead of just the pending writers queue. This would move this
        # from a case-2 RWLock problem to a case-3, ensuring zero starvation of
        # either lock.

        # Condition lock on internal variables
        # noinspection PyUnresolvedReferences
        self.__cond = multiprocessing.Condition()

        # Records the ID of the current writer. If this is None, there isn't
        # currently and entity that has the write lock.
        #: :type: None or int
        self.__writer = None

        # Allows write locks to be reentrant. The write lock is only released
        # when this hits zero during a releaseWrite
        self.__writer_count = 0

        # Writers will be serviced before readers, recorded by this count.
        # Writers serviced in order of request as ordered in this list.
        #: :type: deque of (int, int)
        self.__pending_writers = deque()  # insert linked-list structure here?

        # current processes holding reader locks. Maps the user's ID to the
        # number of times it has reentered reading locks.
        #: :type: dict of (int, int)
        self.__readers = {}

    def acquireRead(self, timeout=None, _id=None):
        """ Acquire a read lock, waiting at most timeout seconds.

        If a timeout is not given, we will block indefinitely. If a negative
        timeout is given, this acquire will act as non-blocking. Raises
        RuntimeError if the timeout expires before lock acquired.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        :param timeout: Optional timeout on the lock acquire time in seconds.
        :type timeout: int or None

        :param _id: Optional identity token.  This should not be used unless you
            know what you are doing.  This is usually the combination of thread
            and process identifiers.
        :type _id: (int, int)

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)
        # print("[DEBUG] acquireRead ID:", me)
        if timeout is not None:
            expire_time = time.time() + timeout
        with self.__cond:
            # If we're the writer, we should also be able to read since reading
            # from other sources is currently locked while we're the writer.
            # Increment our reader reentrant level. Required a matching
            # releaseRead within acquireWrite block.
            if self.__writer == me:
                self.__readers[me] = self.__readers.get(me, 0) + 1
                return
            while True:
                # Only consider granting a read lock if there is currently no
                # writer.
                if not self.__writer:
                    # Increment the reentrant level if we already have a read
                    # lock (including if we're an upgraded reader), else
                    # grant a read lock if there are no pending writers or an
                    # upgraded reader.
                    if self.__readers.get(me):
                        self.__readers[me] += 1
                        return
                    elif not self.__pending_writers:
                        self.__readers[me] = 1
                        return

                if timeout is not None:
                    # noinspection PyUnboundLocalVariable
                    remaining = expire_time - time.time()
                    if remaining <= 0:
                        raise RuntimeError("Timeout expired while waiting for "
                                           "read lock acquire.")
                    self.__cond.wait(remaining)
                else:
                    self.__cond.wait()

    def acquireWrite(self, timeout=None, _id=None):
        """ Acquire a write lock, waiting at most timeout seconds.

        If a timeout is not given, we will block indefinitely. If a negative
        timeout is given, this acquire will act as non-blocking. Raises
        RuntimeError if the timeout expires before lock acquired.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        :param timeout: Optional timeout on the lock acquire time in seconds.
        :type timeout: int or None

        :param _id: Optional identity token.  This should not be used unless you
            know what you are doing.  This is usually the combination of thread
            and process identifiers.
        :type _id: (int, int)

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)
        # print("[DEBUG] acquireWrite ID:", me)
        if timeout is not None:
            expire_time = time.time() + timeout
        with self.__cond:
            # if we're either the writer or an upgraded reader already,
            # increment reentrant level and grant lock.
            if self.__writer == me:
                self.__writer_count += 1
                return

            # Notifying of no read lock upgrade ability at this time.
            elif self.__readers.get(me, False):
                raise ReaderUpdateException("Read lock upgrades not supported "
                                            "at this time.")

            # we're now a normal "pending" writer, no readers will acquire while
            # we are pending.
            else:
                self.__pending_writers.append(me)

            while True:
                # If no readers and no writer, we have clear passage. An
                # upgraded reader would have an entry in __readers if it
                # existed.
                if self.__writer is None and not self.__readers:
                    if self.__pending_writers[0] == me:
                        assert (self.__writer is None and
                                self.__writer_count == 0)
                        self.__writer = self.__pending_writers.popleft()
                        self.__writer_count = 1
                        return

                if timeout is not None:
                    # noinspection PyUnboundLocalVariable
                    remaining = expire_time - time.time()
                    if remaining <= 0:
                        # Timeout expired. Pull ourself from pending deque.
                        # (must be in the queue since in order to get here we
                        # had to have added ourself above).
                        self.__pending_writers.remove(me)
                        raise RuntimeError("Timeout expired while waiting for "
                                           "write lock acquire.")
                    self.__cond.wait(remaining)
                else:
                    self.__cond.wait()

    def releaseRead(self, _id=None):
        """ Release currently acquired read lock.

        Raises RuntimeError if we try to release a read lock when there isn't
        one acquired.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)
        # print("[DEBUG] releaseRead ID:", me)
        with self.__cond:
            # Releasing an inner read lock within an outer write lock. Must have
            # at least one reader level available, else this is an unbound
            # release (no paired acquire).
            if self.__writer == me:
                level = self.__readers.get(me, 0)
                if level > 0:
                    self.__readers[me] -= 1
                    # When out of reader levels, remove us from the map
                    if not self.__readers[me]:
                        del self.__readers[me]
                elif level < 0:
                    raise RuntimeError("Achieved negative reader reentrant "
                                       "level. There's a bug somewhere.")
                else:
                    raise RuntimeError("Attempted a read lock release within "
                                       "a write lock block when no read lock "
                                       "was acquired.")

            # if we're in a read lock
            elif self.__readers.get(me, None):
                # decrement reader count, notify
                self.__readers[me] -= 1

                # If we've released our last read lock, remove ourselves from
                # the mapping of active readers.
                if not self.__readers[me]:
                    del self.__readers[me]

                    # If there are no more readers left, notify-all for those
                    # waiting on this condition.
                    if not self.__readers:
                        self.__cond.notify_all()

            # Erroneous release
            else:
                raise RuntimeError("Attempting release of read lock when one "
                                   "not acquired.")

    def releaseWrite(self, _id=None):
        """ Release currently acquired write lock.

        Raises RuntimeError if we try to release a write lock when there isn't
        one acquired.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)
        # print("[DEBUG] releaseWrite ID:", me)
        with self.__cond:
            # Obviously, only decrement when we own the write lock
            if self.__writer == me:
                self.__writer_count -= 1

                # when we release our initial entry level, clear us from the
                # writer slot
                if not self.__writer_count:
                    # falsify writer flag, notify
                    self.__writer = None
                    self.__cond.notify_all()

            # Erroneous release
            else:
                raise RuntimeError("Attempting release of write lock when one "
                                   "not acquired.")

    def read_lock(self, timeout=None, _id=None):
        """
        Return an object to be used for ``with`` statements that acquires and
        releases read locks. A timeout may be specified, which will be used when
        attempting to acquire our lock.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        :param timeout: Optional timeout on the lock acquire in seconds.
        :type timeout: int

        :param _id: Optional identity token.  This should not be used unless you
            know what you are doing.  This is usually the combination of thread
            and process identifiers.
        :type _id: (int, int)

        :return: New object instance with __enter__ and __exit__ methods
            defined.
        :rtype: object

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)

        # noinspection PyMethodParameters,PyUnusedLocal
        class ReadWithLock (object):

            def __enter__(_self):
                self.acquireRead(timeout, me)

            def __exit__(_self, *args):
                self.releaseRead(me)

        return ReadWithLock()

    def write_lock(self, timeout=None, _id=None):
        """
        Return an object to be used for ``with`` statements that acquires and
        releases write locks. A timeout may be specified, which will be used
        when attempting to acquire our lock.

        NOTE: NO NOT give a value to ``_id`` unless you know what you're doing.
        Otherwise, it should be a tuple of two ints uniquely identifying the
        acquirer of the lock. Subsequent acquisition of locks by the same
        acquirer must provide the same value for correct functionality.
        Otherwise, when _id is None, this is the combination of
        (processID, threadID).

        :param timeout: Optional timeout on the lock acquire in seconds.
        :type timeout: int

        :param _id: Optional identity token.  This should not be used unless you
            know what you are doing.  This is usually the combination of thread
            and process identifiers.
        :type _id: (int, int)

        :return: New object instance with __enter__ and __exit__ methods
            defined.
        :rtype: object

        """
        me = _id or (multiprocessing.current_process().ident,
                     current_thread().ident)

        # noinspection PyMethodParameters,PyUnusedLocal
        class WriteWithLock (object):

            def __enter__(_self):
                self.acquireWrite(timeout, me)

            def __exit__(_self, *args):
                self.releaseWrite(me)

        return WriteWithLock()


class ContextualReadWriteLock (object):
    """
    Case-3 reader-writer lock that does NOT allow read-lock carriers to upgrade
    to a write lock.

    Care should be made to not nest write-lock contexts under an existing read
    or write-lock context, otherwise a dead-lock will occur.

    Source algorithm:
        https://en.wikipedia.org/wiki/Readers%E2%80%93writers_problem#Third_readers-writers_problem
    """

    def __init__(self):
        self._service_lock = multiprocessing.Lock()
        self._resource_lock = multiprocessing.Lock()
        self._reader_count_lock = multiprocessing.Lock()
        self._reader_count = 0

    def read_context(self):
        # noinspection PyMethodParameters
        class ReadLockContext (object):
            def __enter__(_self):
                with self._service_lock:
                    with self._reader_count_lock:
                        if self._reader_count == 0:
                            self._resource_lock.acquire()
                        self._reader_count += 1

            def __exit__(_self, exc_type, exc_val, exc_tb):
                with self._reader_count_lock:
                    self._reader_count -= 1
                    if self._reader_count == 0:
                        self._resource_lock.release()

        return ReadLockContext()

    def write_context(self):
        # noinspection PyMethodParameters
        class WriteLockContext (object):
            def __enter__(_self):
                with self._service_lock:
                    self._resource_lock.acquire()

            def __exit__(_self, exc_type, exc_val, exc_tb):
                self._resource_lock.release()

        return WriteLockContext()



import unittest

import numpy
from six import BytesIO
from six.moves import cPickle, cStringIO

from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement


class TestDescriptorMemoryElement (unittest.TestCase):

    def test_configuration(self):
        default_config = DescriptorMemoryElement.get_default_config()
        self.assertEqual(default_config, {})

        inst1 = DescriptorMemoryElement.from_config(default_config, 'test', 'a')
        self.assertEqual(default_config, inst1.get_config())
        self.assertEqual(inst1.type(), 'test')
        self.assertEqual(inst1.uuid(), 'a')

        # vector-based equality
        inst2 = DescriptorMemoryElement.from_config(inst1.get_config(),
                                                    'test', 'a')
        self.assertEqual(inst1, inst2)

    def test_pickle_dump_load(self):
        # Make a couple descriptors
        v1 = numpy.array([1, 2, 3])
        d1 = DescriptorMemoryElement('test', 0)
        d1.set_vector(v1)

        v2 = numpy.array([4, 5, 6])
        d2 = DescriptorMemoryElement('test', 1)
        d2.set_vector(v2)

        d1_s = cPickle.dumps(d1)
        d2_s = cPickle.dumps(d2)

        # Attempt reconstitution
        d1_r = cPickle.loads(d1_s)
        d2_r = cPickle.loads(d2_s)

        numpy.testing.assert_array_equal(v1, d1_r.vector())
        numpy.testing.assert_array_equal(v2, d2_r.vector())

    def test_set_state_version_1(self):
        # Test support of older state version
        expected_type = 'test-type'
        expected_uid = 'test-uid'
        expected_v = numpy.array([1, 2, 3])
        expected_v_b = BytesIO()
        # noinspection PyTypeChecker
        numpy.save(expected_v_b, expected_v)
        expected_v_dump = expected_v_b.getvalue()

        e = DescriptorMemoryElement(None, None)
        e.__setstate__((expected_type, expected_uid, expected_v_dump))
        self.assertEqual(e.type(), expected_type)
        self.assertEqual(e.uuid(), expected_uid)
        numpy.testing.assert_array_equal(e.vector(), expected_v)

    def test_input_immutability(self):
        # make sure that data stored is not susceptible to shifts in the
        # originating data matrix they were pulled from.

        #
        # Testing this with a single vector
        #
        v = numpy.random.rand(16)
        t = tuple(v.copy())
        d = DescriptorMemoryElement('test', 0)
        d.set_vector(v)
        v[:] = 0
        self.assertTrue((v == 0).all())
        self.assertFalse(sum(t) == 0.)
        numpy.testing.assert_equal(d.vector(), t)

        #
        # Testing with matrix
        #
        m = numpy.random.rand(20, 16)

        v1 = m[3]
        v2 = m[15]
        v3 = m[19]

        # Save truth values of arrays as immutable tuples (copies)
        t1 = tuple(v1.copy())
        t2 = tuple(v2.copy())
        t3 = tuple(v3.copy())

        d1 = DescriptorMemoryElement('test', 1)
        d1.set_vector(v1)
        d2 = DescriptorMemoryElement('test', 2)
        d2.set_vector(v2)
        d3 = DescriptorMemoryElement('test', 3)
        d3.set_vector(v3)

        numpy.testing.assert_equal(v1, d1.vector())
        numpy.testing.assert_equal(v2, d2.vector())
        numpy.testing.assert_equal(v3, d3.vector())

        # Changing the source should not change stored vectors
        m[:, :] = 0.
        self.assertTrue((v1 == 0).all())
        self.assertTrue((v2 == 0).all())
        self.assertTrue((v3 == 0).all())
        self.assertFalse(sum(t1) == 0.)
        self.assertFalse(sum(t2) == 0.)
        self.assertFalse(sum(t3) == 0.)
        numpy.testing.assert_equal(d1.vector(), t1)
        numpy.testing.assert_equal(d2.vector(), t2)
        numpy.testing.assert_equal(d3.vector(), t3)

    def test_output_immutability(self):
        # make sure that data stored is not susceptible to modifications after
        # extraction
        v = numpy.ones(16)
        d = DescriptorMemoryElement('test', 0)
        self.assertFalse(d.has_vector())
        d.set_vector(v)
        r = d.vector()
        r[:] = 0
        self.assertEqual(r.sum(), 0)
        self.assertEqual(d.vector().sum(), 16)

    def test_none_set(self):
        d = DescriptorMemoryElement('test', 0)
        self.assertFalse(d.has_vector())

        d.set_vector(numpy.ones(16))
        self.assertTrue(d.has_vector())
        numpy.testing.assert_equal(d.vector(), numpy.ones(16))

        d.set_vector(None)
        self.assertFalse(d.has_vector())
        self.assertIs(d.vector(), None)



from smqtk.utils import Configurable, plugin, SmqtkObject


# noinspection PyAbstractClass
class SmqtkAlgorithm (SmqtkObject, Configurable, plugin.Pluggable):
    """
    Parent class for all algorithm interfaces.
    """

    @property
    def name(self):
        """
        :return: The name of this class type.
        :rtype: str
        """
        return self.__class__.__name__



"""
LICENCE
-------
Copyright 2013 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.
"""
# -*- coding: utf-8 -*-
from __future__ import print_function
from multiprocessing import current_process
from multiprocessing.managers import (
    all_methods,
    BaseProxy,
    SyncManager
)
import multiprocessing
import multiprocessing.util
import six

import numpy as np
from numpy.core.multiarray import ndarray
# Where the matrix class is depends on version. Either in numpy.core.defmatrix
# or in numpy.matrixlib.defmatrix.
try:
    # noinspection PyUnresolvedReferences
    from numpy.matrixlib.defmatrix import matrix
except ImportError:
    # noinspection PyUnresolvedReferences
    from numpy.core.defmatrix import matrix
    # If this doesn't import, we're either missing version support for the
    # installed version of numpy

from threading import current_thread

from .distance_kernel import DistanceKernel
from .feature_memory import FeatureMemory, FeatureMemoryMap
from . import ReadWriteLock
from .timed_cache import TimedCache


# Create singleton objects if they don't exist a the time of import
if '__singleton_fmm__' not in globals():
    __singleton_fmm__ = FeatureMemoryMap()

if '__singleton_tc__' not in globals():
    __singleton_tc__ = TimedCache()


def get_common_fmm():
    """ Get global singleton FeatureMemoryMap

    :return: Singleton instance of the common FeatureMemoryMap
    :rtype: FeatureMemoryMap

    """
    return __singleton_fmm__


def get_common_tc():
    """ Get global singleton TimedCache

    :return: Singleton instance of common TimedCache
    :rtype: TimedCache

    """
    return __singleton_tc__


#
# Proxy Types and Support
#
# TODO: Check out Pyro as an alternative to multiprocessing.managers

class BaseProxy2 (BaseProxy):
    """
    Intermediate class under BaseProxy in order to fix the resetting of the
    _manager on fork (doesn't need to be? security issues?).
    """

    def __init__(self, token, serializer, manager=None, authkey=None,
                 exposed=None, incref=True):
        super(BaseProxy2, self).__init__(token, serializer, manager, authkey,
                                         exposed, incref)
        # Keep a second manager reference around to survive the base class's
        # after-fork nuke of the _manager attribute.
        self._manager_copy = manager

        multiprocessing.util.register_after_fork(self, BaseProxy2._after_fork2)

    def _after_fork2(self):
        self._manager = self._manager_copy

    def __reduce__(self):
        ret = super(BaseProxy2, self).__reduce__()
        ret[1][-1]['authkey'] = str(self._authkey)
        return ret


def all_properties(obj):
    """
    Return a list of names of non-methods of 'obj'
    """
    noncallables = []
    for name in dir(obj):
        if not hasattr(getattr(obj, name), '__call__'):
            noncallables.append(name)
    return noncallables


# Attributes that should not be overridden in generated proxy objects
__leave_alone_methods = frozenset((
    '__class__', '__delattr__', '__getattribute__', '__init__', '__metaclass__',
    '__new__', '__reduce__', '__reduce_ex__', '__setattr__', '__sizeof__',
    '__subclasshook__',
))


def all_safe_methods(obj):
    """
    Return the tuple of all "safe" method names of the given class. A "safe"
    method name is one that does not belong to the set
    ``__leave_alone_methods``. It is known that overwriting these callable
    methods interferes with the functioning of Proxy objects.

    :param obj: Class to extract callable methods from.
    :type obj: object

    :return: Tuple of safe method names to expose.
    :rtype: tuple of str

    """
    return tuple(set(all_methods(obj)).difference(__leave_alone_methods))


__leave_alone_properties = frozenset((
    '__dict__', '__module__'
))


def all_safe_properties(obj):
    """
    Return the tuple of all "safe" non-callable elements of the given class. A
    "safe" non-callable is one that does not belong to the set
    ``__leave_alone_properties``. It is known that overwriting these
    non-callables interferes with the functioning of Proxy objects.

    :param obj: Class to extract non-callable properties from.
    :type obj: object

    :return: Tuple of safe non-callable property names to expose.
    :rtype: tuple of str

    """
    return tuple(set(all_properties(obj)).difference(__leave_alone_properties))


class ExposedAutoGenMeta (type):

    def __new__(mcs, clsname, bases, dct):
        # look at the _exposed_
        to_expose = set(dct.get('_exposed_', None))
        exposed = set()
        if to_expose is None:
            raise ValueError("Classes using metaclass 'ExposedAutoGenMeta' "
                             "MUST have an ``_exposed_`` iterable defined.")

        # If any of the methods in to_expose are defined in dct, remove them
        # from to_expose and add them to the already exposed set.
        for name in set.intersection(to_expose, dct.keys()):
            if not hasattr(dct[name], '__call__'):
                raise ValueError("Declared an _exposed_ method '%s', but "
                                 "manually defined the same symbol as non-"
                                 "callable." % name)
            to_expose.remove(name)
            exposed.add(name)

        # If the class has a _exposed_properties_ iterable, add __*attr__
        # pass-throughs
        if dct.get('_exposed_properties_', None):
            # remove duplicates if any present
            dct['_exposed_properties_'] = \
                tuple(set(dct['_exposed_properties_']))
            exposed.update(('__getattribute__', '__setattr__', '__delattr__'))

            if '__getattribute__' in to_expose:
                to_expose.remove('__getattribute__')
            if '__getattribute__' in dct:
                print("WARNING: ExposedAutoGenMeta overwriting custom "
                       "``__getattribute__`` in class '%s' in favor of "
                       "property proxy supporting version."
                       % clsname)
            exec("""def __getattr__(self, key):
                if key in self._exposed_properties_:
                    callmethod = object.__getattribute__(self, '_callmethod')
                    return callmethod('__getattribute__', (key,))
                return object.__getattribute__(self, key)""", dct)

            if '__setattr__' in to_expose:
                to_expose.remove('__setattr__')
            if '__setattr__' in dct:
                print("WARNING: ExposedAutoGenMeta overwriting custom "
                       "``__setattr__`` in class '%s' in favor of "
                       "property proxy supporting version."
                       % clsname)
            exec("""def __setattr__(self, key, value):
                if key in self._exposed_properties_:
                    callmethod = object.__getattribute__(self, '_callmethod')
                    return callmethod('__setattr__', (key,))
                return object.__setattr__(self, key, value)""", dct)

            if '__delattr__' in to_expose:
                to_expose.remove('__delattr__')
            if '__delattr__' in dct:
                print("WARNING: ExposedAutoGenMeta overwriting custom "
                       "``__delattr__`` in class '%s' in favor of "
                       "property proxy supporting version."
                       % clsname)
            exec("""def __delattr__(self, key):
                if key in self._exposed_properties_:
                    callmethod = object.__getattribute__(self, '_callmethod')
                    return callmethod('__delattr__', (key,))
                return object.__delattr__(self, key)""", dct)

        # Create default method stamps for remaining methods.
        for method in to_expose:
            exec('''def %s(self, *args, **kwds):
            return self._callmethod(%r, args, kwds)''' % (method, method), dct)
            exposed.add(method)

        # Refresh class var with those methods that have been exposed
        dct['_exposed_'] = tuple(sorted(exposed))

        return super(ExposedAutoGenMeta, mcs).__new__(mcs, clsname, bases, dct)


# Wrapper around numpy ndarray class, providing pass-through functions for class
# functions as of numpy 1.8.0
# TODO: Add``method_to_typeid`` for numpy.array functions that return copies of
#       data.
@six.add_metaclass(ExposedAutoGenMeta)
class ArrayProxy (BaseProxy2):
    _exposed_ = all_safe_methods(ndarray)
    _exposed_properties_ = all_safe_properties(ndarray)
    _method_to_typeid_ = {
        '__iter__': "Iterator"
    }


# Wrapper around numpy matrix class, providing pass-through functions for class
# functions as of numpy 1.8.0.
# TODO: Add``method_to_typeid`` for numpy.matrix functions that return copies of
#       data.
@six.add_metaclass(ExposedAutoGenMeta)
class MatrixProxy (BaseProxy2):
    _exposed_ = all_safe_methods(matrix)
    _exposed_properties_ = all_safe_properties(matrix)
    _method_to_typeid_ = {
        '__iter__': "Iterator"
    }


@six.add_metaclass(ExposedAutoGenMeta)
class RWLockWithProxy (BaseProxy2):
    _exposed_ = ('__enter__', '__exit__')


# noinspection PyPep8Naming
@six.add_metaclass(ExposedAutoGenMeta)
class ReadWriteLockProxy (BaseProxy2):
    _exposed_ = all_safe_methods(ReadWriteLock)
    _method_to_typeid_ = {
        'read_lock':    'rRWLockWith',
        'write_lock':   'rRWLockWith'
    }

    def acquireRead(self, timeout=None, _id=None):
        # Overwriting userID lock uses as the caller of this proxy method.
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('acquireRead', (timeout, me))

    def acquireWrite(self, timeout=None, _id=None):
        # Overwriting userID lock uses as the caller of this proxy method.
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('acquireWrite', (timeout, me))

    def releaseRead(self, _id=None):
        # Overwriting userID lock uses as the caller of this proxy method.
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('releaseRead', (me,))

    def releaseWrite(self, _id=None):
        # Overwriting userID lock uses as the caller of this proxy method.
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('releaseWrite', (me,))

    def read_lock(self, timeout=None, _id=None):
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('read_lock', (timeout, me))

    def write_lock(self, timeout=None, _id=None):
        me = _id or (current_process().ident, current_thread().ident)
        return self._callmethod('write_lock', (timeout, me))


# Return types can't seem to be the same as a type that is creatable, so we use
# a standard 'r' character prefix on return types. This required a manager
# register call below with that name with the option ``create_method=False``.

@six.add_metaclass(ExposedAutoGenMeta)
class DistanceKernelProxy (BaseProxy2):
    _exposed_ = all_safe_methods(DistanceKernel)
    _method_to_typeid_ = {
        'get_lock':             'rReadWriteLock',
        'row_id_map':           'rarray',
        'col_id_map':           'rarray',
        'get_kernel_matrix':    'rmatrix',
        'get_background_ids':   'rarray'
    }


@six.add_metaclass(ExposedAutoGenMeta)
class FeatureMemoryProxy (BaseProxy2):
    _exposed_ = all_safe_methods(FeatureMemory)
    _method_to_typeid_ = {
        '_generate_distance_kernel_matrix': 'rmatrix',

        'get_ids':              'rarray',
        'get_bg_ids':           'rarray',
        'get_feature_matrix':   'rmatrix',
        'get_kernel_matrix':    'rmatrix',
        'get_lock':             'rReadWriteLock',

        'get_distance_kernel':  'rDistanceKernel',
        'get_feature':          'rmatrix',
    }


@six.add_metaclass(ExposedAutoGenMeta)
class FeatureMemoryMapProxy (BaseProxy2):
    _exposed_ = all_safe_methods(FeatureMemoryMap)
    _method_to_typeid_ = {
        'get_feature_memory':   "rFeatureMemory",
        'get_distance_kernel':  "rDistanceKernel",
        'get_feature':          "rmatrix",
    }


@six.add_metaclass(ExposedAutoGenMeta)
class TimedCacheProxy (BaseProxy2):
    _exposed_ = all_safe_methods(TimedCache)
    _method_to_typeid_ = {
        'get_rlock':    'RLock',
        'get_fm':       'rFeatureMemory',
        'get_dk':       'rDistanceKernel',
    }


#
# ProxyManager and Type registration
#
class ProxyManager (SyncManager):
    """
    This class shouldn't be initialized directly, but instead initialized and
    retrieved through the init* and get* functions below.
    """
    pass


# Object-based
ProxyManager.register('array',            np.array,         ArrayProxy)
ProxyManager.register('ndarray',          np.ndarray,       ArrayProxy)
ProxyManager.register('matrix',           np.matrix,        MatrixProxy)
ProxyManager.register('ReadWriteLock',    ReadWriteLock,    ReadWriteLockProxy)
ProxyManager.register('DistanceKernel',   DistanceKernel,   DistanceKernelProxy)
ProxyManager.register('FeatureMemory',    FeatureMemory,    FeatureMemoryProxy)
ProxyManager.register('FeatureMemoryMap', FeatureMemoryMap,
                      FeatureMemoryMapProxy)
# Function based
ProxyManager.register('get_common_fmm',   get_common_fmm,
                      FeatureMemoryMapProxy)
ProxyManager.register('get_common_tc',    get_common_tc,    TimedCacheProxy)
ProxyManager.register("symmetric_dk_from_file",
                      DistanceKernel.construct_symmetric_from_files,
                      DistanceKernelProxy)
ProxyManager.register("asymmetric_dk_from_file",
                      DistanceKernel.construct_asymmetric_from_files,
                      DistanceKernelProxy)

# Return proxy registrations
# - If a function Proxy has a method that should return another proxy, the
#   proxy type that it returns CANNOT have been registered with
#   ``create_method=True``. Reason: the server uses the registered typeid's
#   callable to create the return object
ProxyManager.register('rarray',
                      proxytype=ArrayProxy,
                      create_method=False)
ProxyManager.register("rmatrix",
                      proxytype=MatrixProxy,
                      create_method=False)
# Generated structure from read_lock() and write_lock() methods in ReadWriteLock
ProxyManager.register("rRWLockWith",
                      proxytype=RWLockWithProxy,
                      create_method=False)
ProxyManager.register("rReadWriteLock",
                      proxytype=ReadWriteLockProxy,
                      create_method=False)
ProxyManager.register("rDistanceKernel",
                      proxytype=DistanceKernelProxy,
                      create_method=False)
ProxyManager.register("rFeatureMemory",
                      proxytype=FeatureMemoryProxy,
                      create_method=False)


# Map of ProxyManager instance keyed on the address
#: :type: dict[tuple or str or None, ProxyManager]
__mgr_cache__ = {}


# noinspection PyPep8Naming
def initProxyManagerConnection(address=None, authkey=None,
                               serializer='pickle'):
    """ Initialize ProxyManager connection

    :raises ValueError: The given address already maps to an initialized
        ProxyManager.

    :param address: The address of the ProxyManager server to connect to, or
        None for the local process tree's ProxyManager. When None, the server
        is a child process of this process, or the parent process that first
        created the local server.
    :type address: str or (str, int) or None

    :param authkey: The authentication key for the server we are connecting to.
        This is only used when address is None and this is the first time we
        are getting the feature manager for this process tree (creating the
        connection).
    :type authkey: None or str

    :param serializer: Transport serialization method.  This can be 'pickle' or
        'xmlrpclib' (see ``multiprocessing.managers.listener_client``).

    """
    global __mgr_cache__
    if address in __mgr_cache__:
        raise ValueError("The given address '%s' already maps to an "
                         "initialized ProxyManager!"
                         % str(address))
    # Nope, I'm pretty sure that constructor exists in the base class...
    # noinspection PyArgumentList
    __mgr_cache__[address] = ProxyManager(address, authkey, serializer)

    # if address is None, then we are initializing a local process tree server.
    # i.e. we need to fork and start it
    try:
        if address is None:
            __mgr_cache__[address].start()
        else:
            __mgr_cache__[address].connect()
    # If an error occurred, rollback what we just added
    except Exception:
        del __mgr_cache__[address]
        raise


def removeProxyManagerConnection(address=None):
    """ Shutdown and remove an initialized ProxyManager connection
    :raises KeyError: if the given address is not associated with to an active
        manager.
    :param address: The address of the ProxyManager connection to remove from
        the active mapping.
    :type address: str or (str, int) or None
    """
    global __mgr_cache__
    #: :type: ProxyManager
    if hasattr(__mgr_cache__[address], 'shutdown'):
        __mgr_cache__[address].shutdown()
    del __mgr_cache__[address]


# noinspection PyPep8Naming
def getProxyManager(address=None):
    """
    Get the ProxyManager instance for the given address from the cache. If the
    address is None, returns the ProxyManager for the current process tree
    (may have been initialized on a parent process).

    :raises KeyError: When the given address has not been initialized and is not
        present in the cache.

    :param address: The address of the ProxyManager connection to retrieve
        from the active mapping.
    :type address: str or (str, int) or None

    :return: Singleton feature_manager for the given address.
    :rtype: ProxyManager

    """
    return __mgr_cache__[address]



from __future__ import division, print_function
import unittest

import mock

from smqtk.representation.descriptor_index import DescriptorIndex


class DummyDescriptorIndex (DescriptorIndex):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def get_descriptor(self, uuid):
        pass

    def get_many_descriptors(self, uuids):
        pass

    def iterkeys(self):
        pass

    def iteritems(self):
        pass

    def iterdescriptors(self):
        pass

    def remove_many_descriptors(self, uuids):
        pass

    def has_descriptor(self, uuid):
        pass

    def add_many_descriptors(self, descriptors):
        pass

    def count(self):
        pass

    def clear(self):
        pass

    def remove_descriptor(self, uuid):
        pass

    def add_descriptor(self, descriptor):
        pass


class TestDescriptorIndexAbstract (unittest.TestCase):

    def test_len(self):
        di = DummyDescriptorIndex()
        di.count = mock.Mock(return_value=100)
        self.assertEqual(len(di), 100)
        di.count.assert_called_once_with()

    def test_get_item(self):
        di = DummyDescriptorIndex()
        di.get_descriptor = mock.Mock(return_value='foo')
        self.assertEqual(di['some_key'], 'foo')
        di.get_descriptor.assert_called_once_with('some_key')

    def test_del_item(self):
        di = DummyDescriptorIndex()
        di.remove_descriptor = mock.Mock()

        del di['foo']
        di.remove_descriptor.assert_called_once_with('foo')

    def test_iter(self):
        # Iterating over a DescriptorIndex should yield the descriptor elements
        di = DummyDescriptorIndex()

        def dumb_iterator():
            for _i in range(3):
                yield _i

        di.iterdescriptors = mock.Mock(side_effect=dumb_iterator)

        for i, v in enumerate(iter(di)):
            self.assertEqual(i, v)
        self.assertEqual(list(di), [0, 1, 2])
        self.assertEqual(tuple(di), (0, 1, 2))
        self.assertEqual(di.iterdescriptors.call_count, 3)



import threading
import time
import six

import six

from smqtk.representation import DescriptorElement
from smqtk.representation import DescriptorElementFactory
from smqtk.representation import get_descriptor_element_impls


class CachingDescriptorElement (DescriptorElement):

    @classmethod
    def is_usable(cls):
        """
        This implementation has no direct dependencies of its own.
        :rtype: bool
        """
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        aside from the first two assumed positional arguments, turning those
        argument names into configuration dictionary keys.
        If any of those arguments have defaults, we will add those values into
        the configuration dictionary appropriately.
        The dictionary returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(CachingDescriptorElement, cls).get_default_config()

        # Nested DescriptorElementFactory configuration
        if c['wrapped_element_factory'] is None:
            # Have to make this configuration in such a way that we don't
            # include ourselves in the list of nestable classes else an infinite
            # recursion will occur.

            de_impls = get_descriptor_element_impls()
            # Remove ourselves
            del de_impls[cls.__name__]

            # Construct config block DescriptorElementFactory wants
            c_def = {"type": None}
            for label, de_cls in six.iteritems(de_impls):
                # noinspection PyUnresolvedReferences
                c_def[label] = de_cls.get_default_config()
            c['wrapped_element_factory'] = c_def
        else:
            c['wrapped_element_factory'] = \
                c['wrapped_element_factory'].get_config()

        return c

    @classmethod
    def from_config(cls, config_dict, type_str, uuid, merge_default=True):
        # convert factory configuration
        config_dict['wrapped_element_factory'] = \
            DescriptorElementFactory.from_config(
                config_dict['wrapped_element_factory']
            )

        return super(CachingDescriptorElement, cls).from_config(
            config_dict, type_str, uuid, merge_default
        )

    def __init__(self, type_str, uuid, wrapped_element_factory,
                 cache_expiration_timeout=1.0, poll_interval=0.1):
        """
        Initialize a new caching wrapper descriptor element.

        This implementation is intended to wrap another DescriptorElement type,
        adding a timed caching layer on top of

        :raises AssertionError: Cache expiration seconds was not a positive
            value.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        :param wrapped_element_factory: DescriptorElementFactory to produce
            DescriptorElement instances of the wrapped type.
        :type wrapped_element_factory:
            smqtk.representation.DescriptorElementFactory

        :param cache_expiration_timeout: Timeout in seconds for accessed
            descriptors to be cached. If this is non-zero, a monitoring thread
            will be launched in order to track the timeout. The thread will be
            shutdown after timeout. This value must be positive.

            If this is positive infinity, then the cache never expires. This
            also means that the cache will not be updated if the vector that
            would be returned from the wrapped element ever changes.
        :type cache_expiration_timeout: None | float

        :param poll_interval: How often to check if the cache has expired.
        :type poll_interval: float

        """
        super(CachingDescriptorElement, self).__init__(type_str, uuid)

        self.wrapped_element_factory = wrapped_element_factory
        self.cache_expiration_timeout = float(cache_expiration_timeout)
        self.poll_interval = poll_interval

        assert cache_expiration_timeout > 0, \
            "Cache expiration timeout was not positive."

        self._d_elem = self.wrapped_element_factory \
                           .new_descriptor(self.type(), self.uuid())
        # self._log.debug("Caching descriptor element instance of type '%s'",
        #                 self._d_elem.__class__.__name__)

        # Attributes for timed caching with threads
        self.cache_v = None  # Numpy ndarray if there is a current cache
        self.cache_last_access = None  # UNIX timestamp
        self.cache_lock = threading.RLock()
        #: :type: threading.Thread
        self.cache_thread = None  # the expiry monitor if there is a cached v

    def __del__(self):
        """
        Release vector cache
        """
        with self.cache_lock:
            # Should cause dependent thread to terminate gracefully
            self.cache_v = None
        if self.cache_thread and \
                self.cache_thread.ident != threading.currentThread():
            self.cache_thread.join()

    def __getstate__(self):
        state = super(CachingDescriptorElement, self).__getstate__()
        state.update({
            "wrapped_element_factory": self.wrapped_element_factory,
            "cache_expiration_timeout": self.cache_expiration_timeout,
            "poll_interval": self.poll_interval
        })
        return state

    def __setstate__(self, c):
        # Support older state version
        if 'type' in c:
            self._type_label = c['type']
            self._uuid = c['uuid']
        else:
            super(CachingDescriptorElement, self).__setstate__(c)

        self.wrapped_element_factory = c['wrapped_element_factory']
        self.cache_expiration_timeout = c['cache_expiration_timeout']
        self.poll_interval = c['poll_interval']

        # Initializing local cache variables that were un-pickle-able
        self._d_elem = self.wrapped_element_factory\
                           .new_descriptor(self.type(), self.uuid())
        self.cache_v = None
        self.cache_last_access = None
        self.cache_lock = threading.RLock()
        self.cache_thread = None

    def get_config(self):
        return {
            "wrapped_element_factory": self.wrapped_element_factory,
            "cache_expiration_timeout": self.cache_expiration_timeout,
            "poll_interval": self.poll_interval,
        }

    def has_vector(self):
        """
        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool
        """
        return self.vector() is not None

    def vector(self):
        """
        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.core.multiarray.ndarray or None
        """
        with self.cache_lock:
            v = self.cache_v

            # If no cache, attempt to populate it
            if v is None:

                # No cache currently, attempt fetch from wrapped elem
                # self._log.debug("Getting vector from nested descriptor "
                #                 "element")
                v = self._d_elem.vector()
                # self._log.debug("Vector received: %s", v)

                if v is not None:
                    # Clean-up old thread if there was one
                    if self.cache_thread:
                        # cache_v is None at this point
                        # self._log.debug("Joining old monitor thread")
                        self.cache_lock.release()
                        self.cache_thread.join()
                        self.cache_lock.acquire()

                    # vector in elem; set in cache; start monitor thread
                    self.cache_v = v
                    self.cache_thread = threading.Thread(
                        target=CachingDescriptorElement
                        .thread_monitor_cache_expiration,
                        args=(self,),
                        # verbose=self._log.getEffectiveLevel()<=logging.DEBUG,
                    )
                    # self._log.debug("Spawning cache monitor thread")
                    self.cache_thread.start()

            else:
                # Currently have a cache
                assert self.cache_thread is not None, \
                    "Have a cache, but no monitor thread."

            self.cache_last_access = time.time()

        return v

    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.core.multiarray.ndarray

        :returns: Self.
        :rtype: CachingDescriptorElement

        """
        # set source vector and set as current cache
        with self.cache_lock:
            # self._log.debug("Setting in source element")
            self._d_elem.set_vector(new_vec)

            # Only start monitor when we're given a cache-able vector and the
            #   cache is currently not occupied.
            if new_vec is not None and self.cache_v is None:
                if self.cache_thread:
                    # cache_thread not alive at this point
                    # self._log.debug("Joining old monitor thread")
                    self.cache_lock.release()
                    self.cache_thread.join()
                    self.cache_lock.acquire()

                # need to start new monitor thread
                self.cache_thread = threading.Thread(
                    target=CachingDescriptorElement
                    .thread_monitor_cache_expiration,
                    args=(self,),
                    # verbose=self._log.getEffectiveLevel() <= logging.DEBUG,
                )
                # self._log.debug("Spawning cache monitor thread")
                self.cache_thread.start()

            # Update cache vector and access time
            self.cache_v = new_vec
            self.cache_last_access = time.time()
        return self

    @staticmethod
    def thread_monitor_cache_expiration(elem):
        """
        Monitor wrapper element's cache expiration

        :param elem: Wrapper element instance for monitoring
        :type elem: CachingDescriptorElement

        """
        # log = logging.getLogger(__name__)
        #
        # # noinspection PyProtectedMember
        # log_header = '[{type:s}, {uuid:s}, {elem:s}]'.format(**{
        #     "type": elem.type(),
        #     'uuid': elem.uuid(),
        #     'elem': elem.wrapped_element_factory._d_type,
        # })

        expired = False
        while not expired:
            time.sleep(elem.poll_interval)
            with elem.cache_lock:
                t = time.time()
                # log.debug("%s Checking cache cache expiration "
                #           "[now = %f | last access = %f | timeout = %f]",
                #           log_header, t, elem.cache_last_access,
                #           elem.cache_expiration_timeout)
                if t - elem.cache_last_access >= elem.cache_expiration_timeout:
                    elem.cache_v = None
                    expired = True
                    # log.debug("%s Cache expired", log_header)
                elif elem.cache_v is None:
                    expired = True
                    # log.debug("%s Cache was invalidated for us", log_header)
        # log.debug("%s Monitor thread exiting", log_header)


# Disabling this implementation for the moment because it needs to be rethought
DESCRIPTOR_ELEMENT_CLASS = None



from __future__ import print_function
import six

import mock
import os
import unittest

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation.data_element import from_uri
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.tests import TEST_DATA_DIR


class TestDataFileElement (unittest.TestCase):

    def test_init_filepath_abs(self):
        fp = '/foo.txt'
        d = DataFileElement(fp)
        self.assertEqual(d._filepath, fp)

    def test_init_relFilepath_normal(self):
        # relative paths should be stored as given within the element
        fp = 'foo.txt'
        d = DataFileElement(fp)
        self.assertEqual(d._filepath, fp)

    def test_content_type(self):
        d = DataFileElement('foo.txt')
        self.assertEqual(d.content_type(), 'text/plain')

    def test_content_type_explicit_type(self):
        ex_type = 'image/png'
        d = DataFileElement('foo.txt', explicit_mimetype=ex_type)
        self.assertEqual(d.content_type(), ex_type)

    @mock.patch('smqtk.representation.data_element.DataElement.write_temp')
    def test_writeTempOverride(self, mock_DataElement_wt):
        # no manual directory, should return the base filepath
        expected_filepath = '/path/to/file.txt'
        d = DataFileElement(expected_filepath)
        fp = d.write_temp()

        self.assertFalse(mock_DataElement_wt.called)
        self.assertEqual(expected_filepath, fp)

    @mock.patch('smqtk.representation.data_element.DataElement.write_temp')
    def test_writeTempOverride_sameDir(self, mock_DataElement_wt):
        expected_filepath = '/path/to/file.txt'
        target_dir = '/path/to'

        d = DataFileElement(expected_filepath)
        fp = d.write_temp(temp_dir=target_dir)

        self.assertFalse(mock_DataElement_wt.called)
        self.assertEqual(fp, expected_filepath)

    @mock.patch('smqtk.representation.data_element.DataElement.write_temp')
    def test_writeTempOverride_diffDir(self, mock_DataElement_wt):
        """
        Test that adding ``temp_dir`` parameter triggers call to parent class
        """
        source_filepath = '/path/to/file.png'
        target_dir = '/some/other/dir'

        d = DataFileElement(source_filepath)

        # Should call parent class write_temp since target is not the same dir
        # that the source file is in.
        mock_DataElement_wt.return_value = 'expected'
        v = d.write_temp(temp_dir=target_dir)
        self.assertEqual(v, 'expected')
        mock_DataElement_wt.assert_called_with(target_dir)

    def test_cleanTemp(self):
        # a write temp and clean temp should not affect original file
        source_file = os.path.join(TEST_DATA_DIR, 'test_file.dat')
        self.assertTrue(os.path.isfile(source_file))
        d = DataFileElement(source_file)
        d.write_temp()
        self.assertEqual(len(d._temp_filepath_stack), 0)
        d.clean_temp()
        self.assertTrue(os.path.isfile(source_file))

    def test_fromConfig(self):
        fp = os.path.join(TEST_DATA_DIR, "Lenna.png")
        c = {
            "filepath": fp
        }
        #: :type: DataFileElement
        df = DataFileElement.from_config(c)
        self.assertEqual(df._filepath, fp)

    def test_toConfig(self):
        fp = os.path.join(TEST_DATA_DIR, "Lenna.png")
        df = DataFileElement(fp)
        c = df.get_config()
        self.assertEqual(c['filepath'], fp)

    def test_configuration(self):
        fp = os.path.join(TEST_DATA_DIR, "Lenna.png")
        default_config = DataFileElement.get_default_config()
        self.assertEqual(default_config,
                         {'filepath': None, 'readonly': False,
                          'explicit_mimetype': None})

        default_config['filepath'] = fp
        inst1 = DataFileElement.from_config(default_config)
        self.assertEqual(default_config, inst1.get_config())

        inst2 = DataFileElement.from_config(inst1.get_config())
        self.assertEqual(inst1, inst2)

    def test_repr(self):
        e = DataFileElement('foo')
        self.assertEqual(repr(e),
                         "DataFileElement{filepath: foo, readonly: False, "
                         "explicit_mimetype: None}")

        e = DataFileElement('bar', readonly=True)
        self.assertEqual(repr(e),
                         "DataFileElement{filepath: bar, readonly: True, "
                         "explicit_mimetype: None}")

        e = DataFileElement('baz', readonly=True, explicit_mimetype='some/type')
        self.assertEqual(repr(e),
                         "DataFileElement{filepath: baz, readonly: True, "
                         "explicit_mimetype: some/type}")

    def test_from_uri_invalid_uri_empty(self):
        # Given empty string
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            ''
        )

    def test_from_uri_invalid_uri_malformed_rel_directory(self):
        # URI malformed: relative path trailing slash (directory)
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            "some/rel/path/dir/"
        )

    def test_from_uri_invalid_uri_malformed_abs_directory(self):
        # URI malformed: absolute path trailing slash (directory)
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            "/abs/path/dir/"
        )

    def test_from_uri_invalid_uri_malformed_bad_header(self):
        # URI malformed: file:// malformed

        # Missing colon
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            "file///some/file/somewhere.txt"
        )

        # file misspelled
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            "fle:///some/file/somewhere.txt"
        )

    def test_from_uri_invalid_uri_malformed_header_rel_path(self):
        # URL malformed: file:// not given ABS path
        self.assertRaises(
            InvalidUriError,
            DataFileElement.from_uri,
            "file://some/rel/path.txt"
        )

    # noinspection PyUnresolvedReferences
    def test_from_uri(self):
        # will be absolute path
        test_file_path = os.path.join(TEST_DATA_DIR, "test_file.dat")
        print("Test file path:", test_file_path)

        e = DataFileElement.from_uri(test_file_path)
        self.assertIsInstance(e, DataFileElement)
        self.assertEqual(e._filepath, test_file_path)
        self.assertEqual(e.get_bytes(), six.b(''))

        e = DataFileElement.from_uri('file://' + test_file_path)
        self.assertIsInstance(e, DataFileElement)
        self.assertEqual(e._filepath, test_file_path)
        self.assertEqual(e.get_bytes(), six.b(''))

    # noinspection PyUnresolvedReferences
    def test_from_uri_plugin_level(self):
        # will be absolute path
        test_file_path = os.path.join(TEST_DATA_DIR, "test_file.dat")
        print("Test file path:", test_file_path)

        e = from_uri(test_file_path)
        self.assertIsInstance(e, DataFileElement)
        self.assertEqual(e._filepath, test_file_path)
        self.assertEqual(e.get_bytes(), six.b(''))

        e = from_uri('file://' + test_file_path)
        self.assertIsInstance(e, DataFileElement)
        self.assertEqual(e._filepath, test_file_path)
        self.assertEqual(e.get_bytes(), six.b(''))

    def test_is_empty_file_not_exists(self):
        e = DataFileElement('/no/exists')
        self.assertTrue(e.is_empty())

    def test_is_empty_file_zero_data(self):
        e = DataFileElement(os.path.join(TEST_DATA_DIR, 'test_file.dat'))
        self.assertTrue(e.is_empty())

    def test_is_empty_file_has_data(self):
        e = DataFileElement(os.path.join(TEST_DATA_DIR, 'Lenna.png'))
        self.assertFalse(e.is_empty())

    def test_get_bytes_no_file(self):
        e = DataFileElement("/not/a/valid/path.txt", readonly=True)
        # We currently expect, in the case where the filepath doesn't exist, to
        # get the same bytes as if the file existed and were empty.
        self.assertEqual(e.get_bytes(), six.b(""))
        # read-only status should have no effect.
        e = DataFileElement("/not/a/valid/path.txt", readonly=True)
        self.assertEqual(e.get_bytes(), six.b(""))

    def test_get_bytes(self):
        # Test with a known real file.
        test_file_path = os.path.join(TEST_DATA_DIR, 'text_file')
        e = DataFileElement(test_file_path)
        self.assertEqual(e.get_bytes(), six.b("Some text content.\n"))

    def test_writable_readonly_false(self):
        e = DataFileElement('foo')
        self.assertTrue(e.writable())

        e = DataFileElement('foo', False)
        self.assertTrue(e.writable())

        e = DataFileElement('foo', readonly=False)
        self.assertTrue(e.writable())

    def test_writable_readonly_true(self):
        e = DataFileElement('foo', True)
        self.assertFalse(e.writable())

        e = DataFileElement('foo', readonly=True)
        self.assertFalse(e.writable())

    @mock.patch('smqtk.representation.data_element.file_element'
                '.safe_file_write')
    def test_set_bytes_writable(self, m_sfw):
        # Using a relative filepath
        test_path = 'foo'
        test_bytes = six.b('test string of bytes')

        e = DataFileElement(test_path)
        e.set_bytes(test_bytes)

        # File write function should be called
        m_sfw.assert_called_once_with(test_path, test_bytes)

    def test_set_bytes_readonly(self):
        e = DataFileElement('foo', readonly=True)
        self.assertRaises(
            ReadOnlyError,
            e.set_bytes,
            six.b('some bytes')
        )



from smqtk.representation import \
    SmqtkRepresentation, \
    get_descriptor_element_impls
from smqtk.utils.plugin import make_config
from smqtk.utils import merge_dict


__author__ = "paul.tunison@kitware.com"


class DescriptorElementFactory (SmqtkRepresentation):
    """
    Factory class for producing DescriptorElement instances of a specified type
    and configuration.
    """

    def __init__(self, d_type, type_config):
        """
        Initialize the factory to produce DescriptorElement instances of the
        given type and configuration.

        :param d_type: Type of descriptor element this factory should produce.
        :type d_type: type

        :param type_config: Initialization parameter dictionary that should
            contain all additional construction parameters for the provided type
            except for the expected `type_str` and `uuid` arguments that should
            be the first and second positional arguments respectively.
        :type type_config: dict

        """
        #: :type: smqtk.representation.DescriptorElement
        self._d_type = d_type
        self._d_type_config = type_config

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        return make_config(get_descriptor_element_impls())

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless and instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: DescriptorElementFactory

        """
        if merge_default:
            merged_config = cls.get_default_config()
            merge_dict(merged_config, config_dict)
            config_dict = merged_config

        return DescriptorElementFactory(
            get_descriptor_element_impls()[config_dict['type']],
            config_dict[config_dict['type']]
        )

    def get_config(self):
        d_type_name = self._d_type.__name__
        return {
            'type': d_type_name,
            d_type_name: self._d_type_config,
        }

    def new_descriptor(self, type_str, uuid):
        """
        Create a new DescriptorElement instance of the configured implementation

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: UUID to associate with the descriptor
        :type uuid: collections.Hashable

        :return: New DescriptorElement instance
        :rtype: smqtk.representation.DescriptorElement

        """
        return self._d_type.from_config(self._d_type_config, type_str, uuid)

    def __call__(self, type_str, uuid):
        """
        Create a new DescriptorElement instance of the configured implementation

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: UUID to associate with the descriptor
        :type uuid: collections.Hashable

        :return: New DescriptorElement instance
        :rtype: smqtk.representation.DescriptorElement

        """
        return self.new_descriptor(type_str, uuid)



import imageio
import logging
import os
import shutil

import six

from smqtk.utils import file_utils, video_utils
from smqtk.utils.mimetype import get_mimetypes


MIMETYPES = get_mimetypes()


class PreviewCache (object):
    """
    Create and cache saved located of preview images for data elements.
    """

    # Preview generation methods based on content type
    # - types can either be specific types or just type classes (i.e. "image" or
    #   "video"), which are fall-backs when a data element's specific content
    #   type is not found in the mapping.
    # - This should be populated with methods that take two arguments, the first
    #   being a DataElement instance, and the second being the directory to
    #   place generated files under. These methods should return the full path
    #   to the generated preview image.
    #: :type: dict[collections.Hashable, collections.Callable]
    PREVIEW_GEN_METHOD = {}

    @property
    def _log(self):
        return logging.getLogger('.'.join([self.__module__,
                                           self.__class__.__name__]))

    def __init__(self, cache_dir):
        """
        :param cache_dir: Directory to cache preview image elements into.
        :type cache_dir: str
        """
        self._cache_dir = os.path.abspath(os.path.expanduser(cache_dir))
        # Cache of preview images for data elements encountered.
        #: :type: dict[collections.Hashable, str]
        self._preview_cache = {}
        self._video_work_dir = os.path.join(cache_dir, 'tmp_video_work')

    def __del__(self):
        """
        Cleanup after ourselves.
        """
        for fp in six.itervalues(self._preview_cache):
            os.remove(fp)

    def get_preview_image(self, elem):
        """
        Get the filepath to the preview image for the given data element.

        :raises ValueError: Do not know how to generate a preview image for the
            given element's content type.

        :param elem: Data element to generate a preview image for.
        :type elem: smqtk.representation.DataElement

        :return: Path to the preview image for the given data element.
        :rtype: str

        """
        if elem.uuid() in self._preview_cache:
            return self._preview_cache[elem.uuid()]

        # else, generate preview image based on content type / content class
        if elem.content_type() in self.PREVIEW_GEN_METHOD:
            self._log.debug("Generating preview image based on content type: "
                            "%s", elem.content_type)
            file_utils.safe_create_dir(self._cache_dir)
            fp = self.PREVIEW_GEN_METHOD[elem.content_type()](self, elem,
                                                              self._cache_dir)
        else:
            content_class = elem.content_type().split('/', 1)[0]
            if content_class in self.PREVIEW_GEN_METHOD:
                self._log.debug("Generating preview image based on content "
                                "class: %s", content_class)
                file_utils.safe_create_dir(self._cache_dir)
                fp = self.PREVIEW_GEN_METHOD[content_class](self, elem,
                                                            self._cache_dir)
            else:
                raise ValueError("No preview generation method for the data "
                                 "element provided, of content type '%s'."
                                 % elem.content_type())
        self._preview_cache[elem.uuid()] = fp
        return fp

    # noinspection PyMethodMayBeStatic
    def gen_image_preview(self, elem, output_dir):
        """
        Copy temporary image to specified output filepath.

        :param elem: Data element to get the preview image for.
        :type elem: smqtk.representation.DataElement

        :param output_dir: Directory to save generated image to.
        :type output_dir: str

        """
        output_fp = os.path.join(
            output_dir,
            "%s%s" % (str(elem.uuid()),
                      MIMETYPES.guess_extension(elem.content_type()))
        )
        if not os.path.isfile(output_fp):
            tmp_img_fp = elem.write_temp()
            shutil.copy(tmp_img_fp, output_fp)
            elem.clean_temp()
        return output_fp

    def gen_video_preview(self, elem, output_dir):
        """
        Copy temporary image to specified output filepath.

        :param elem: Data element to get the preview image for.
        :type elem: smqtk.representation.DataElement

        :param output_dir: Directory to save generated image to.
        :type output_dir: str

        """
        output_fp = os.path.join(output_dir,
                                 "%s.gif" % elem.uuid())
        if not os.path.isfile(output_fp):
            tmp_vid_fp = elem.write_temp()
            interval = 0.5  # ~2fps gif
            fm = video_utils.ffmpeg_extract_frame_map(
                self._video_work_dir, tmp_vid_fp,
                second_interval=interval
            )
            img_arrays = []
            for frm_num in sorted(fm.keys()):
                img_arrays.append(imageio.imread(fm[frm_num]))
            imageio.mimwrite(output_fp, img_arrays, duration=interval)
            elem.clean_temp()
        return output_fp


PreviewCache.PREVIEW_GEN_METHOD = {
    "image": PreviewCache.gen_image_preview,
    "video": PreviewCache.gen_video_preview,
}



"""
Utilities for interacting with Girder
"""

import datetime

import requests

from smqtk.utils import SmqtkObject
from smqtk.utils.url import url_join


class GirderTokenManager (SmqtkObject):
    """
    Helper class to manage storage and update of a Girder authentication token.

    This is better than using the girder_client module in that this class will
    request a new token with the current one expires.
    """

    def __init__(self, api_root='http://localhost:8080/api/v1', api_key=None):
        """
        Initialize a new token manager

        :param api_root: Girder API root URL
        :type api_root: str

        :param api_key: Optional API key to request token with. Otherwise an
        :type api_key:

        """
        self._api_root = api_root
        self._api_key = api_key
        self._token = None
        self._expiration = None

    @property
    def api_root(self):
        return self._api_root

    @property
    def api_key(self):
        return self._api_key

    @property
    def token(self):
        return self._token

    @property
    def expiration(self):
        return self._expiration

    @staticmethod
    def _parse_expiration_timestamp(ts):
        """
        Parse datetime instance from the given expiration timestamp string.
        Currently ignores timezone, but asserts that its '+00:00' (UTC).

        :raises AssertionError: Timestamp did not have a UTC timezone specifier
            attacked to the end.

        :param ts: Token expiration timestamp string.
        :type ts: unicode

        :return: Datetime instance parsed from timestamp.
        :rtype: datetime.datetime

        """
        # example:
        #   2017-02-25T14:59:48.333777+00:00
        # Ignoring timezone at the end for now and asserting that its always
        # '+00:00' (UTC)
        ts, tz = ts[:-6], ts[-6:]
        assert tz == "+00:00", "Expiration UTC timezone assumption broken, " \
                               "received: '%s'" % tz
        dt = datetime.datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%f")
        return dt

    def _request_token(self):
        """
        Request an authentication token given an API key.

        Assumes ``'self.api_key`` is defined and a valid API key value.

        :raises AssertionError: Expiration timestamp did not have a UTC timezone
            specifier attacked to the end.

        :return: token string and expiration timestamp
        :rtype: str, datetime.datetime

        """
        self._log.debug("Requesting new authorization token.")
        if self._api_key:
            r = requests.post(
                url_join(self._api_root, 'api_key/token'),
                data={'key': self._api_key}
            )
            r.raise_for_status()
            token = r.json()['authToken']['token']
            expires = r.json()['authToken']['expires']
        else:
            r = requests.get(
                url_join(self._api_root, 'token/session')
            )
            r.raise_for_status()
            token = r.json()['token']
            expires = r.json()['expires']
        return token, self._parse_expiration_timestamp(expires)

    def _check_token_expiration(self):
        """
        Check if our current auth token has expired or if we don't have one yet.
        If so, request a new one. Only does anything if we have an api_key set.

        Assumes ``'self.api_key`` is defined and a valid API key value.

        :raises AssertionError: Expiration timestamp did not have a UTC timezone
            specifier attacked to the end.
        """
        if (self._token is None
                or self.token_expiration <= datetime.datetime.now()):
            self._log.debug("No or expired token")
            self._token, self.token_expiration = self._request_token()

    def get_token(self):
        self._check_token_expiration()
        return self._token

    def get_requests_header(self):
        """
        :return: The token authorization header if we have an api_key set.
            Otherwise returns None.
        :rtype: None | dict
        """
        self._check_token_expiration()
        return {'Girder-Token': self._token}



"""
Add a number of Girder hosted files to a dataset.

Files are discovered by recursively looking under any folders and items
provided. Explicit file IDs may be additionally provided. We also take in an
optional new-line separated file list of folder, item and file IDs.

Tool Configuration Parameters
-----------------------------

    girder_api_root
        Root URL where Girder is being hosted.

    api_key
        Optional user API key to provide additional permissions if target data
        is not public.

    api_query_batch
        Number of elements to query for at a time when expanding folder and item
        contents. A batch size of 0 means all elements are queried for at once.

    dataset_insert_batch_size
        An optional number of elements to batch before adding to the configured
        data-set. If this is 0 or None, we will collect all elements before
        adding then to the configured data-set.

TODO: Add support for searching collections

"""

import collections
import logging

import requests

from smqtk.representation import get_data_set_impls
from smqtk.representation.data_element.girder import GirderDataElement
from smqtk.utils import bin_utils, plugin
from smqtk.utils.girder import GirderTokenManager
from smqtk.utils.url import url_join


################################################################################
# Component Functions

def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)
    g_girder = parser.add_argument_group('Girder References')

    g_girder.add_argument('-F', '--folder',
                          nargs='*', default=[], metavar='FOLDER_ID',
                          help='Specify specific folder IDs')
    g_girder.add_argument('--folder-list', metavar='PATH',
                          help='Path to a new-line separated file of folder '
                               'IDs')

    g_girder.add_argument('-i', '--item',
                          nargs='*', default=[], metavar='ITEM_ID',
                          help='Specify specific item IDs')
    g_girder.add_argument('--item-list', metavar='PATH',
                          help='Path to a new-line separated file of item IDs')

    g_girder.add_argument('-f', '--file',
                          nargs='*', default=[], metavar='FILE_ID',
                          help='Specify specific file IDs')
    g_girder.add_argument('--file-list', metavar='PATH',
                          help='Path to a new-line separated file of file IDs')

    return parser


def default_config():
    return {
        'tool': {
            'girder_api_root': 'http://localhost:8080/api/v1',
            'api_key': None,
            'api_query_batch': 1000,
            'dataset_insert_batch_size': None,
        },
        'plugins': {
            'data_set': plugin.make_config(get_data_set_impls()),
        }
    }


def iterate_query(q, batch_size):
    """

    :param q: function taking an offset and limit parameter, returning an
        iterable of elements to yield.
    :type q: (int, int) -> __generator

    :param batch_size: Number of elements to get from a single query.
    :type batch_size: int | None

    :return: Elements yielded by nested function
    :rtype: __generator

    """
    query_again = True
    y_total = 0
    while query_again:
        y_this_query = 0
        for r in q(y_total, batch_size):
            yield r
            y_this_query += 1
        query_again = batch_size > 0 and (y_this_query == batch_size)
        y_total += y_this_query


def get_folder_subfolders(api_root, folder_id, tm, batch_size):
    """
    Iterate folder sub-folder IDs

    :type api_root: str
    :type folder_id: str
    :type tm: GirderTokenManager
    :type batch_size: int
    :rtype: __generator[str]
    """
    def q(offset, limit):
        r = requests.get(url_join(api_root, 'folder'),
                         params={'parentType': 'folder',
                                 'parentId': folder_id,
                                 'offset': offset,
                                 'limit': limit},
                         headers=tm.get_requests_header())
        r.raise_for_status()
        for f_model in r.json():
            yield f_model['_id']

    return iterate_query(q, batch_size)


def get_folder_items(api_root, folder_id, tm, batch_size):
    """
    Iterate folder item IDs

    :type api_root: str
    :type folder_id: str
    :type tm: GirderTokenManager
    :type batch_size: int
    :rtype: __generator[str]
    """
    def q(offset, limit):
        r = requests.get(url_join(api_root, 'item'),
                         params={'folderId': folder_id,
                                 'offset': offset,
                                 'limit': limit},
                         headers=tm.get_requests_header())
        r.raise_for_status()
        for i_model in r.json():
            yield i_model['_id']

    return iterate_query(q, batch_size)


def get_item_files(api_root, item_id, tm, batch_size):
    """
    Iterate item file IDs and content type

    :type api_root: str
    :type item_id: str
    :type tm: GirderTokenManager
    :type batch_size: int
    :rtype: __generator[str]
    """
    def q(offset, limit):
        r = requests.get(url_join(api_root, 'item', item_id, 'files'),
                         params={'offset': offset,
                                 'limit': limit},
                         headers=tm.get_requests_header())
        r.raise_for_status()
        for file_model in r.json():
            yield file_model['_id'], file_model['mimeType']

    return iterate_query(q, batch_size)


def find_girder_files(api_root, folder_ids, item_ids, file_ids,
                      api_key=None, query_batch=None):
    """
    Query girder for file IDs nested under folders and items, and yield
    GirderDataElement instances for each child file discovered.

    We first find nested folder and items under given folder IDs, then files
    under discovered items, finally yielding GirderDataElements for file IDs
    discovered.

    Data elements yielded are in order of file IDs given, to files under items
    given, to files under items discovered in folders in a breadth-first order.

    :param api_root: Root URL of the girder instance to call.
    :type api_root: str

    :param folder_ids: Iterable of Girder folder IDs to recursively collect file
        elements from.
    :type folder_ids: collection.Iterable[str]

    :param item_ids: Iterable of Girder item IDs to collect file elements from.
    :type item_ids: collection.Iterable[str]

    :param file_ids: Iterable of Girder file IDs to make elements of.
    :type file_ids: collection.Iterable[str]

    :param api_key: Optional user API key to use for authentication when
        accessing private data.
    :type api_key: None | str

    :param query_batch: Number of elements to query for at a time when expanding
        folder and item contents. A batch size of 0 means all elements are
        queried for at once.
    :type query_batch: int

    :return: Generator yielding GirderDataElement instances.
    :rtype: __generator[GirderDataElement]

    """
    log = logging.getLogger(__name__)
    tm = GirderTokenManager(api_root, api_key)
    # Get the token once before parallel requests
    tm.get_token()

    if query_batch is None:
        query_batch = 0

    # Could also do something with multi-threading/processing

    log.info("Yielding elements from file IDs")
    file_fifo = collections.deque(file_ids)
    while file_fifo:  # Just file IDs
        file_id = file_fifo.popleft()
        log.debug('-f %s', file_id)
        e = GirderDataElement(file_id, api_root, api_key)
        e.token_manager = tm  # because we already made one
        yield e

    log.info("Collecting files from items")
    item_fifo = collections.deque(item_ids)
    while item_fifo:
        item_id = item_fifo.popleft()
        log.debug('-i %s', item_id)
        for file_id, ct in get_item_files(api_root, item_id, tm, query_batch):
            log.debug('   -f %s', file_id)
            e = GirderDataElement(file_id, api_root, api_key)
            e._content_type = ct
            e.token_manager = tm
            yield e

    # Collect items from folders, then files from items.
    log.info("Collecting items from folders")
    folder_fifo = collections.deque(folder_ids)
    while folder_fifo:
        folder_id = folder_fifo.popleft()
        log.debug("-F %s", folder_id)

        # Find sub-folders
        folder_fifo.extend(get_folder_subfolders(api_root, folder_id, tm,
                                                 query_batch))

        for item_id in get_folder_items(api_root, folder_id, tm, query_batch):
            log.debug('   -i %s', item_id)
            for file_id, ct in get_item_files(api_root, item_id, tm,
                                              query_batch):
                log.debug('      -f %s (%s)', file_id, ct)
                e = GirderDataElement(file_id, api_root, api_key)
                e._content_type = ct
                e.token_manager = tm
                yield e


################################################################################
# Main

def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    api_root = config['tool']['girder_api_root']
    api_key = config['tool']['api_key']
    api_query_batch = config['tool']['api_query_batch']
    insert_batch_size = config['tool']['dataset_insert_batch_size']

    # Collect N folder/item/file references on CL and any files referenced.
    #: :type: list[str]
    ids_folder = args.folder
    #: :type: list[str]
    ids_item = args.item
    #: :type: list[str]
    ids_file = args.file

    if args.folder_list:
        with open(args.folder_list) as f:
            ids_folder.extend([fid.strip() for fid in f])
    if args.item_list:
        with open(args.item_list) as f:
            ids_item.extend([iid.strip() for iid in f])
    if args.file_list:
        with open(args.file_list) as f:
            ids_file.extend([fid.strip() for fid in f])

    #: :type: smqtk.representation.DataSet
    data_set = plugin.from_plugin_config(config['plugins']['data_set'],
                                         get_data_set_impls())

    batch = collections.deque()
    rps = [0]*7
    for e in find_girder_files(api_root, ids_folder, ids_item, ids_file,
                               api_key, api_query_batch):
        batch.append(e)
        if insert_batch_size and len(batch) >= insert_batch_size:
            data_set.add_data(*batch)
            batch.clear()
        bin_utils.report_progress(log.info, rps, 1.0)

    if batch:
        data_set.add_data(*batch)

    log.info('Done')


################################################################################
# Entry-point

if __name__ == '__main__':
    main()



import mock
import unittest

import numpy
from six.moves import cPickle

from smqtk.representation.descriptor_element.local_elements import \
    DescriptorFileElement


class TestDescriptorFileElement (unittest.TestCase):

    def test_configuration1(self):
        default_config = DescriptorFileElement.get_default_config()
        self.assertEqual(default_config,
                         {
                             'save_dir': None,
                             'subdir_split': None,
                         })

        default_config['save_dir'] = '/some/path/somewhere'
        default_config['subdir_split'] = 4

        #: :type: DescriptorFileElement
        inst1 = DescriptorFileElement.from_config(default_config,
                                                  'test', 'abcd')
        self.assertEqual(default_config, inst1.get_config())
        self.assertEqual(inst1._save_dir, '/some/path/somewhere')
        self.assertEqual(inst1._subdir_split, 4)

        # vector-based equality
        inst2 = DescriptorFileElement.from_config(inst1.get_config(),
                                                  'test', 'abcd')
        self.assertEqual(inst1, inst2)

    def test_vec_filepath_generation(self):
        d = DescriptorFileElement('test', 'abcd', '/base', 4)
        self.assertEqual(d._vec_filepath,
                         '/base/a/b/c/test.abcd.vector.npy')

        d = DescriptorFileElement('test', 'abcd', '/base', 2)
        self.assertEqual(d._vec_filepath,
                         '/base/ab/test.abcd.vector.npy')

        d = DescriptorFileElement('test', 'abcd', '/base', 1)
        self.assertEqual(d._vec_filepath,
                         '/base/test.abcd.vector.npy')

        d = DescriptorFileElement('test', 'abcd', '/base', 0)
        self.assertEqual(d._vec_filepath,
                         '/base/test.abcd.vector.npy')

        d = DescriptorFileElement('test', 'abcd', '/base')
        self.assertEqual(d._vec_filepath,
                         '/base/test.abcd.vector.npy')

    def test_serialization(self):
        # Test that an instance can be serialized and deserialized via pickle
        # successfully.
        ex_type = 'test'
        ex_uid = 12345
        ex_save_dir = 'some-dir'
        ex_split = 5
        e1 = DescriptorFileElement(ex_type, ex_uid, ex_save_dir, ex_split)

        # pickle dump and load into a new copy
        #: :type: DescriptorFileElement
        e2 = cPickle.loads(cPickle.dumps(e1))
        # Make sure the two have the smme attributes, including base descriptor
        # element things.
        self.assertEqual(e1.type(), e2.type())
        self.assertEqual(e1.uuid(), e2.uuid())
        self.assertEqual(e1._save_dir, e2._save_dir)
        self.assertEqual(e1._subdir_split, e2._subdir_split)
        self.assertEqual(e1._vec_filepath, e2._vec_filepath)

    @mock.patch('smqtk.representation.descriptor_element.local_elements'
                '.numpy.save')
    @mock.patch('smqtk.representation.descriptor_element.local_elements'
                '.file_utils.safe_create_dir')
    def test_vector_set(self, mock_scd, mock_save):
        d = DescriptorFileElement('test', 1234, '/base', 4)
        self.assertEqual(d._vec_filepath,
                         '/base/1/2/3/test.1234.vector.npy')

        v = numpy.zeros(16)
        d.set_vector(v)
        mock_scd.assert_called_with('/base/1/2/3')
        mock_save.assert_called_with('/base/1/2/3/test.1234.vector.npy', v)

    @mock.patch('smqtk.representation.descriptor_element.local_elements'
                '.numpy.load')
    def test_vector_get(self, mock_load):
        d = DescriptorFileElement('test', 1234, '/base', 4)
        self.assertFalse(d.has_vector())
        self.assertIs(d.vector(), None)

        d.has_vector = mock.Mock(return_value=True)
        self.assertTrue(d.has_vector())
        v = numpy.zeros(16)
        mock_load.return_value = v
        numpy.testing.assert_equal(d.vector(), v)



# -*- coding: utf-8 -*-

import math


__all__ = ["sieve_of_eratosthenes", "prime_factors", "factors", "factor_pairs"]


# Set of known prime values
_soe_prime_cache = [2, 3]
# Map of not-prime values and the increment after that value where the next
# not-prime value may be found. This should only contain keys greater-than
# the max(_prime_cache) value for finding further prime values.
_soe_not_prime_map = {9: 3}
# The next value to continue iterating from.
_soe_c = 5
# TODO: Lock these globals when in use.


def sieve_of_eratosthenes(N):
    """
    Return an ascending ordered list of prime numbers up to the given value,
    generated via the sieve of eratosthenes method.

    TODO: Add starting offset so as to be able to get primes within a range.

    :param N: Value to get prime values less-than-or-equal this value.
    :type N: int | float | str

    :return: List of prime integer values <= N.
    :rtype: list[int]

    """
    # TODO: Optimization - all primes after 2 are odd, so special case when
    #           N == 2, else loop over and deal only with odd numbers.
    global _soe_c

    # in case we get N as a string or something
    N = float(N)

    if _soe_c > N:
        # We have already computed all primes up to N, so return cache-filled
        # list.
        return [p for p in _soe_prime_cache if p <= N]

    # We need to find more prime values, so start with the least known
    # not-prime and sieve our way up.  We maintain that _soe_c is always odd
    # since no even value can be prime.
    while _soe_c <= N:
        if _soe_c in _soe_not_prime_map:
            # Get increment for the current non-prime value and record the next
            # odd non-prime value, keeping this increment. This may increment
            # past N in order to main a correct state in the cache structure.
            # Since we skip even values, _soe_c and its increment value will
            # always be odd. Furthermore, two odd numbers added together will
            # always be even so in order to the next odd not-prime value we need
            # to add in increments of 2 times the base increment value.
            incr = _soe_not_prime_map[_soe_c]
            incr2 = incr * 2
            nnp = _soe_c + incr2
            # Continue finding next divisible value that is not divisible by a
            # higher prime.
            while nnp in _soe_not_prime_map:
                nnp += incr2
            _soe_not_prime_map[nnp] = incr
            # Remove _soe_c from map as it has been considered now.
            del _soe_not_prime_map[_soe_c]
        else:
            # _soe_c is a prime value: record it and add square of _soe_c to
            # not-prime-map (square of the prime is the first value not
            # attainable by multiplying preceding prime value).
            _soe_prime_cache.append(_soe_c)
            _soe_not_prime_map[_soe_c*_soe_c] = _soe_c
        _soe_c += 2

    # since we had to extend it, return a copy of the cache
    return list(_soe_prime_cache)


def prime_factors(N):
    """
    Returns ordered ascending prime factors of N.

    If a floating point value is passed, we cast it to an integer.

    :param N: Value to get the prime factors list of.
    :type N: int

    :returns: List of prime factors in ascending order.
    :rtype: list[int]

    """
    sqrt_N = math.sqrt(N)
    primes = sieve_of_eratosthenes(sqrt_N)

    # prime factorize N
    p_factors = []
    remaining = int(N)
    for p in primes:
        # Short-cut loop exit when we know we cant factorize any more
        if remaining == 1:
            break
        # factory out
        while remaining > 1 and ((remaining % p) == 0):
            p_factors.append(p)
            remaining /= p

    if remaining != 1:
        p_factors.append(remaining)

    return p_factors


def factors(N):
    """
    Return divisors/factors of N.

    :param N: Value to get the factors of.
    :type N: int

    :returns: Set of factors for value N.
    :rtype: set[int]

    """
    pf = prime_factors(N)
    # noinspection PySetFunctionToLiteral
    ftors = set([1])

    # Compute product of all primes
    for p in pf:
        new_factors = set()
        # multiply the current prime factor with all current factors in order to
        # iteratively build up component factors of N
        for e in ftors:
            new_factors.add(e * p)
        ftors.update(new_factors)

    # using sorted as we want to return a list, and directly converting a set
    # to a list causes things to be out of sequential order.
    return ftors


def factor_pairs(N):
    """
    Return a list of factor pairs of N, whose individual products produce N.

    :param N: Value to get the factor pairs for.
    :type N: int

    :returns: List of factor pairs ordered by minimum factor value.
    :rtype: list[(int, int)]

    """
    # Order factors in order to match outside edges of shrinking bounds.
    ftors = sorted(factors(N))
    # Step indices from the edges towards the center, pairing the edges to
    # produce factor pairs.
    i = 0
    j = len(ftors) - 1
    pairs = []
    while i <= j:
        pairs.append((ftors[i], ftors[j]))
        i += 1
        j -= 1
    return pairs



import abc
import numpy
import os

from smqtk.representation import SmqtkRepresentation
from smqtk.utils import plugin
from smqtk.utils import merge_dict


__author__ = "paul.tunison@kitware.com"


class DescriptorElement (SmqtkRepresentation, plugin.Pluggable):
    """
    Abstract descriptor vector container.

    This structure supports implementations that cache descriptor vectors on a
    per-UUID basis.

    UUIDs must maintain unique-ness when transformed into a string.

    Descriptor element equality based on shared descriptor type and vector
    equality. Two descriptor vectors that are generated by different types of
    descriptor generator should not be considered the same (though, this may be
    up for discussion).

    Stored vectors should be effectively immutable.

    """

    def __init__(self, type_str, uuid):
        """
        Initialize a new descriptor element.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        """
        super(DescriptorElement, self).__init__()

        self._type_label = type_str
        self._uuid = uuid

    def __hash__(self):
        return hash(self.uuid())

    def __eq__(self, other):
        if isinstance(other, DescriptorElement):
            return numpy.array_equal(self.vector(), other.vector())
        return False

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s{type: %s, uuid: %s}" % (self.__class__.__name__, self.type(),
                                           self.uuid())

    def __getstate__(self):
        return {
            "_type_label": self._type_label,
            "_uuid": self._uuid,
        }

    def __setstate__(self, state):
        self._type_label = state['_type_label']
        self._uuid = state['_uuid']

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        aside from the first two assumed positional arguments, turning those
        argument names into configuration dictionary keys.
        If any of those arguments have defaults, we will add those values into
        the configuration dictionary appropriately.
        The dictionary returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        # similar to parent impl, except we remove the ``type_str`` and ``uuid``
        # configuration parameters as they are to be specified at runtime.
        dc = super(DescriptorElement, cls).get_default_config()
        # These parameters must be specified at construction time.
        del dc['type_str'], dc['uuid']
        return dc

    # noinspection PyMethodOverriding
    @classmethod
    def from_config(cls, config_dict, type_str, uuid, merge_default=True):
        """
        Instantiate a new instance of this class given the desired type, uuid,
        and JSON-compliant configuration dictionary.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: DescriptorElement

        """
        c = {}
        merge_dict(c, config_dict)
        c['type_str'] = type_str
        c['uuid'] = uuid
        return super(DescriptorElement, cls).from_config(c, merge_default)

    def uuid(self):
        """
        :return: Unique ID for this vector.
        :rtype: collections.Hashable
        """
        return self._uuid

    def type(self):
        """
        :return: Type label type of the DescriptorGenerator that generated this
            vector.
        :rtype: str
        """
        return self._type_label

    ###
    # Abstract methods
    #

    @abc.abstractmethod
    def has_vector(self):
        """
        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool
        """

    @abc.abstractmethod
    def vector(self):
        """
        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.ndarray or None
        """

    @abc.abstractmethod
    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.ndarray

        :returns: Self.
        :rtype: DescriptorMemoryElement

        """


from ._io import *


def get_descriptor_element_impls(reload_modules=False):
    """
    Discover and return discovered ``DescriptorElement`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable
          ``DESCRIPTOR_ELEMENT_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``DESCRIPTOR_ELEMENT_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``DescriptorElement``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "DESCRIPTOR_ELEMENT_PATH"
    helper_var = "DESCRIPTOR_ELEMENT_CLASS"
    return plugin.get_plugins(__name__, this_dir, env_var, helper_var,
                              DescriptorElement, reload_modules)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import collections
from copy import deepcopy
import ctypes
import logging
import os
import tempfile

import numpy
import numpy.linalg
import scipy.stats
import six
from six.moves import cPickle

from smqtk.algorithms import SupervisedClassifier
from smqtk.representation.data_element import from_uri
from smqtk.representation.descriptor_element import elements_to_matrix

try:
    import svm
    import svmutil
except ImportError:
    svm = None
    svmutil = None


class LibSvmClassifier (SupervisedClassifier):
    """
    Classifier that uses libSVM for support-vector machine functionality.

    **Note**: *If pickled without a having model file paths configured, this
    implementation will write out temporary files upon pickling and loading.
    This is required because the model instance is not transportable via
    serialization due to libSVM being an external C library.*
    """

    @classmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        This implementation required the libSVM python bindings to be installed
        and loadable.

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        return None not in {svm, svmutil}

    # noinspection PyDefaultArgument
    def __init__(self, svm_model_uri=None, svm_label_map_uri=None,
                 train_params={
                     '-s': 0,  # C-SVC, assumed default if not provided
                     '-t': 0,  # linear kernel
                     '-b': 1,  # enable probability estimates
                     '-c': 2,  # SVM parameter C
                     # '-g': 0.0078125,  # initial gamma (1 / 128)
                 },
                 normalize=None,
                 ):
        """
        Initialize the classifier with an empty or existing model.

        Model file paths are optional. If they are given and the file(s) exist,
        we will load them. If they do not, we treat the path(s) as the output
        path(s) for saving a model after calling ``train``. If this is None
        (default), no model is loaded nor output via training, thus any model
        trained will only exist in memory during the lifetime of this instance.

        :param svm_model_uri: Path to the libSVM model file.
        :type svm_model_uri: None | str

        :param svm_label_map_uri: Path to the pickle file containing this
            model's output labels.
        :type svm_label_map_uri: None | str

        :param train_params: SVM parameters used for training. See libSVM
            documentation for parameter flags and values.
        :type train_params: dict[basestring, int|float]

        :param normalize: Normalize input vectors to training and
            classification methods using ``numpy.linalg.norm``. This may either
            be  ``None``, disabling normalization, or any valid value that
            could be passed to the ``ord`` parameter in ``numpy.linalg.norm``
            for 1D arrays. This is ``None`` by default (no normalization).
        :type normalize: None | int | float | str

        """
        super(LibSvmClassifier, self).__init__()

        self.svm_model_uri = svm_model_uri
        self.svm_label_map_uri = svm_label_map_uri

        # Elements will be None if input URI is None
        #: :type: None | smqtk.representation.DataElement
        self.svm_model_elem = \
            svm_model_uri and from_uri(svm_model_uri)
        #: :type: None | smqtk.representation.DataElement
        self.svm_label_map_elem = \
            svm_label_map_uri and from_uri(svm_label_map_uri)

        self.train_params = train_params
        self.normalize = normalize
        # Validate normalization parameter by trying it on a random vector
        if normalize is not None:
            self._norm_vector(numpy.random.rand(8))

        # generated parameters
        #: :type: svm.svm_model
        self.svm_model = None
        # dictionary mapping SVM integer labels to semantic labels
        #: :type: dict[int, collections.Hashable]
        self.svm_label_map = None

        self._reload_model()

    def __getstate__(self):
        # If we don't have a model, or if we have one but its not being saved
        # to files.
        if not self.has_model() or (self.svm_model_uri is not None and
                                    self.svm_label_map_uri is not None):
            return self.get_config()
        else:
            self._log.debug("Saving model to temp file for pickling")
            fd, fp = tempfile.mkstemp()
            try:
                os.close(fd)

                state = self.get_config()
                state['__LOCAL__'] = True
                state['__LOCAL_LABELS__'] = self.svm_label_map

                fp_bytes = fp.encode('utf8')
                svmutil.svm_save_model(fp_bytes, self.svm_model)
                with open(fp, 'rb') as model_f:
                    state['__LOCAL_MODEL__'] = model_f.read()

                return state
            finally:
                os.remove(fp)

    def __setstate__(self, state):
        self.svm_model_uri = state['svm_model_uri']
        self.svm_label_map_uri = state['svm_model_uri']
        self.train_params = state['train_params']
        self.normalize = state['normalize']

        self.svm_model_elem = \
            self.svm_model_uri and from_uri(self.svm_model_uri)
        self.svm_label_map_elem = \
            self.svm_label_map_uri and from_uri(self.svm_label_map_uri)

        # C libraries/pointers don't survive across processes.
        if '__LOCAL__' in state:
            fd, fp = tempfile.mkstemp()
            try:
                os.close(fd)

                self.svm_label_map = state['__LOCAL_LABELS__']

                # write model to file, then load via libSVM
                with open(fp, 'wb') as model_f:
                    model_f.write(state['__LOCAL_MODEL__'])

                fp_bytes = fp.encode('utf8')
                self.svm_model = svmutil.svm_load_model(fp_bytes)

            finally:
                os.remove(fp)
        else:
            self.svm_model = None
            self._reload_model()

    def _reload_model(self):
        """
        Reload SVM model from configured file path.
        """
        if self.svm_model_elem and not self.svm_model_elem.is_empty():
            svm_model_tmp_fp = self.svm_model_elem.write_temp()
            self.svm_model = svmutil.svm_load_model(svm_model_tmp_fp)
            self.svm_model_elem.clean_temp()

        if self.svm_label_map_elem and not self.svm_label_map_elem.is_empty():
            self.svm_label_map = \
                cPickle.loads(self.svm_label_map_elem.get_bytes())

    @staticmethod
    def _gen_param_string(params):
        """
        Make a single string out of a parameters dictionary
        """
        return ' '.join((str(k) + ' ' + str(v)
                         for k, v in six.iteritems(params)))

    def _norm_vector(self, v):
        """
        Class standard array normalization. Normalized along max dimension (a=0
        for a 1D array, a=1 for a 2D array, etc.).

        :param v: Vector to normalize
        :type v: numpy.ndarray

        :return: Returns the normalized version of input array ``v``.
        :rtype: numpy.ndarray

        """
        if self.normalize is not None:
            n = numpy.linalg.norm(v, self.normalize, v.ndim - 1,
                                  keepdims=True)
            # replace 0's with 1's, preventing div-by-zero
            n[n == 0.] = 1.
            return v / n

        # Normalization off
        return v

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            "svm_model_uri": self.svm_model_uri,
            "svm_label_map_uri": self.svm_label_map_uri,
            "train_params": self.train_params,
            "normalize": self.normalize,
        }

    def has_model(self):
        """
        :return: If this instance currently has a model loaded. If no model is
            present, classification of descriptors cannot happen.
        :rtype: bool
        """
        return None not in (self.svm_model, self.svm_label_map)

    def _train(self, class_examples, **extra_params):
        """
        Internal method that trains the classifier implementation.

        This method is called after checking that there is not already a model
        trained, thus it can be assumed that no model currently exists.

        The class labels will have already been checked before entering this
        method, so it can be assumed that the ``class_examples`` will container
        at least two classes.

        :param class_examples: Dictionary mapping class labels to iterables of
            DescriptorElement training examples.
        :type class_examples: dict[collections.Hashable,
                 collections.Iterable[smqtk.representation.DescriptorElement]]

        :param extra_params: Dictionary with extra parameters for training.
            This is not used by this implementation.
        :type extra_params: None | dict[basestring, object]

        """

        # Offset from 0 for positive class labels to use
        # - not using label of 0 because we think libSVM wants positive labels
        CLASS_LABEL_OFFSET = 1

        # Stuff for debug reporting
        etm_ri = None
        param_debug = {'-q': ''}
        if self._log.getEffectiveLevel() <= logging.DEBUG:
            etm_ri = 1.0
            param_debug = {}

        # Form libSVM problem input values
        self._log.debug("Formatting problem input")
        train_labels = []
        train_vectors = []
        train_group_sizes = []  # number of examples per class
        self.svm_label_map = {}
        # Making SVM label assignment deterministic to alphabetic order
        for i, l in enumerate(sorted(class_examples), CLASS_LABEL_OFFSET):
            # Map integer SVM label to semantic label
            self.svm_label_map[i] = l

            self._log.debug('-- class %d (%s)', i, l)
            # requires a sequence, so making the iterable ``g`` a tuple
            g = class_examples[l]
            if not isinstance(g, collections.Sequence):
                self._log.debug('   (expanding iterable into sequence)')
                g = tuple(g)

            train_group_sizes.append(float(len(g)))
            x = elements_to_matrix(g, report_interval=etm_ri)
            x = self._norm_vector(x)
            train_labels.extend([i] * x.shape[0])
            train_vectors.extend(x.tolist())
            del g, x

        assert len(train_labels) == len(train_vectors), \
            "Count mismatch between parallel labels and descriptor vectors" \
            "being sent to libSVM (%d != %d)" \
            % (len(train_labels), len(train_vectors))

        self._log.debug("Forming train params")
        #: :type: dict
        params = deepcopy(self.train_params)
        params.update(param_debug)
        # Calculating class weights if set to C-SVC type SVM
        if '-s' not in params or int(params['-s']) == 0:
            # (john.moeller): The weighting should probably be the geometric
            # mean of the number of examples over the classes divided by the
            # number of examples for the current class.
            gmean = scipy.stats.gmean(train_group_sizes)
            for i, n in enumerate(train_group_sizes, CLASS_LABEL_OFFSET):
                w = gmean / n
                params['-w' + str(i)] = w
                self._log.debug("-- class '%s' weight: %s",
                                self.svm_label_map[i], w)

        self._log.debug("Making parameters obj")
        svm_params = svmutil.svm_parameter(self._gen_param_string(params))
        self._log.debug("Creating SVM problem")
        svm_problem = svm.svm_problem(train_labels, train_vectors)
        del train_vectors
        self._log.debug("Training SVM model")
        self.svm_model = svmutil.svm_train(svm_problem, svm_params)
        self._log.debug("Training SVM model -- Done")

        if self.svm_label_map_elem and self.svm_label_map_elem.writable():
            self._log.debug("saving labels to element (%s)",
                            self.svm_label_map_elem)
            self.svm_label_map_elem.set_bytes(
                cPickle.dumps(self.svm_label_map, -1)
            )
        if self.svm_model_elem and self.svm_model_elem.writable():
            self._log.debug("saving model to element (%s)",
                            self.svm_model_elem)
            # LibSvm I/O only works with filepaths, thus the need for an
            # intermediate temporary file.
            fd, fp = tempfile.mkstemp()
            try:
                svmutil.svm_save_model(fp, self.svm_model)
                # Use the file descriptor to create the file object.
                # This avoids reopening the file and will automatically
                # close the file descriptor on exiting the with block.
                # fdopen() is required because in Python 2 open() does
                # not accept a file descriptor.
                with os.fdopen(fd, 'rb') as f:
                    self.svm_model_elem.set_bytes(f.read())
            finally:
                os.remove(fp)

    def get_labels(self):
        """
        Get the sequence of class labels that this classifier can classify
        descriptors into. This includes the negative label.

        :return: Sequence of possible classifier labels.
        :rtype: collections.Sequence[collections.Hashable]

        :raises RuntimeError: No model loaded.

        """
        if not self.has_model():
            raise RuntimeError("No model loaded")
        return list(self.svm_label_map.values())

    def _classify(self, d):
        """
        Internal method that constructs the label-to-confidence map (dict) for
        a given DescriptorElement.

        The passed descriptor element is guaranteed to have a vector to
        extract. It is not extracted yet due to the philosophy of waiting
        until the vector is immediately needed. This moment is thus determined
        by the implementing algorithm.

        :param d: DescriptorElement containing the vector to classify.
        :type d: smqtk.representation.DescriptorElement

        :raises RuntimeError: Could not perform classification for some reason
            (see message in raised exception).

        :return: Dictionary mapping trained labels to classification confidence
            values
        :rtype: dict[collections.Hashable, float]

        """
        if not self.has_model():
            raise RuntimeError("No SVM model present for classification")

        # Get and normalize vector
        v = d.vector().astype(float)
        v = self._norm_vector(v)
        v, idx = svm.gen_svm_nodearray(v.tolist())

        # Effectively reproducing the body of svmutil.svm_predict in order to
        # simplify and get around excessive prints
        svm_type = self.svm_model.get_svm_type()
        nr_class = self.svm_model.get_nr_class()
        c = dict((l, 0.) for l in self.get_labels())

        if self.svm_model.is_probability_model():
            # noinspection PyUnresolvedReferences
            if svm_type in [svm.NU_SVR, svm.EPSILON_SVR]:
                nr_class = 0
            # noinspection PyCallingNonCallable,PyTypeChecker
            prob_estimates = (ctypes.c_double * nr_class)()
            svm.libsvm.svm_predict_probability(self.svm_model, v,
                                               prob_estimates)
            # Update dict
            for l, p in zip(self.svm_model.get_labels(),
                            prob_estimates[:nr_class]):
                c[self.svm_label_map[l]] = p
        else:
            # noinspection PyUnresolvedReferences
            if svm_type in (svm.ONE_CLASS, svm.EPSILON_SVR, svm.NU_SVC):
                nr_classifier = 1
            else:
                nr_classifier = nr_class * (nr_class - 1) // 2
            # noinspection PyCallingNonCallable,PyTypeChecker
            dec_values = (ctypes.c_double * nr_classifier)()
            label = svm.libsvm.svm_predict_values(self.svm_model, v,
                                                  dec_values)
            # Update dict
            c[self.svm_label_map[label]] = 1.

        assert len(c) == len(self.svm_label_map)
        return c


# Explicitly declare to avoid silly warning about ignoring parent abstract
# class.
CLASSIFIER_CLASS = LibSvmClassifier



from smqtk.utils import Configurable, SmqtkObject


#  noinspection PyAbstractClass
class SmqtkRepresentation (SmqtkObject, Configurable):
    """
    Interface for data representation interfaces and implementations.

    Data should be serializable, so this interface adds abstract methods for
    serializing and de-serializing SMQTK data representation instances.

    """
    # TODO(paul.tunison): Add serialization abstract method signatures here.
    # - Could start with just requiring implementing sub-classes to
    #   ``__getstate__`` and ``__setstate__`` methods required for pickle
    #   interface.



import flask


__author__ = 'paul.tunison@kitware.com'


class StaticDirectoryHost (flask.Blueprint):
    """
    Module that will host a given directory to the given URL prefix (relative to
    the parent module's prefix).

    Instances of this class will have nothing set to their static URL path, as a
    blank string is used. Please reference the URL prefix value.

    """

    def __init__(self, name, static_dir, url_prefix):
        # make sure URL prefix starts with a slash
        if not url_prefix.startswith('/'):
            url_prefix = '/' + url_prefix

        super(StaticDirectoryHost, self).__init__(name, __name__,
                                                  static_folder=static_dir,
                                                  static_url_path="",
                                                  url_prefix=url_prefix)



import argparse
import json
import logging
import logging.handlers
import os
import sys
import threading
import time
import warnings

from smqtk.utils import merge_dict, SmqtkObject


def initialize_logging(logger, stream_level=logging.WARNING,
                       output_filepath=None, file_level=None):
    """
    Standard logging initialization.

    :param logger: Logger instance to initialize
    :type logger: logging.Logger

    :param stream_level: Logging level to set for the stderr stream formatter.
    :type stream_level: int

    :param output_filepath: Output logging from the given logger to the provided
        file path. Currently, we log to that file indefinitely, i.e. no
        rollover. Rollover may be added in the future if the need arises.
    :type output_filepath: str

    :param file_level: Logging level to output to the file. This the same as the
        stream level by default.

    """
    log_formatter = logging.Formatter(
        "%(levelname)7s - %(asctime)s - %(name)s.%(funcName)s - %(message)s"
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(log_formatter)
    stream_handler.setLevel(stream_level)
    logger.addHandler(stream_handler)

    if output_filepath:
        # TODO: Setup rotating part of the handler?
        file_handler = logging.handlers.RotatingFileHandler(output_filepath,
                                                            mode='w',
                                                            delay=1)
        file_handler.setFormatter(log_formatter)
        file_handler.setLevel(file_level or stream_level)
        logger.addHandler(file_handler)

    # Because there are two levels checked before a logging message is emitted:
    #   * the logging object's level
    #   * The stream handlers level
    logger.setLevel(min(stream_level, file_level or stream_level))


def load_config(config_path, defaults=None):
    """
    Load the JSON configuration dictionary from the specified filepath.

    If the given path does not point to a valid file, we return an empty
    dictionary or the default dictionary if one was provided, returning False
    as our second return argument.

    :param config_path: Path to the (valid) JSON configuration file.
    :type config_path: str

    :param defaults: Optional default configuration dictionary to merge loaded
        configuration into. If provided, it will be modified in place.
    :type defaults: dict | None

    :return: The result configuration dictionary and if we successfully loaded
        a JSON dictionary from the given filepath.
    :rtype: (dict, bool)

    """
    if defaults is None:
        defaults = {}
    loaded = False
    if config_path and os.path.isfile(config_path):
        with open(config_path) as cf:
            merge_dict(defaults, json.load(cf))
            loaded = True
    return defaults, loaded


def output_config(output_path, config_dict, log=None, overwrite=False,
                  error_rc=1):
    """
    If a valid output configuration path is provided, we output the given
    configuration dictionary as JSON or error if the file already exists (when
    overwrite is False) or cannot be written. We exit the program as long as
    ``output_path`` was given a value, with a return code of 0 if the file was
    written successfully, or the supplied return code (default of 1) if the
    write failed.

    Specified error return code cannot be 0, which is reserved for successful
    operation.

    :raises ValueError: If the given error return code is 0.

    :param output_path: Path to write the configuration file to.
    :type output_path: str

    :param config_dict: Configuration dictionary containing JSON-compliant
        values.
    :type config_dict: dict

    :param overwrite: If we should clobber any existing file at the specified
        path. We exit with the error code if this is false and a file exists at
        ``output_path``.
    :type overwrite: bool

    :param error_rc: Custom integer error return code to use instead of 1.
    ;type error_rc: int

    :param log: Optionally logging instance. Otherwise we use a local one.
    :type log: logging.Logger

    """
    error_rc = int(error_rc)
    if error_rc == 0:
        raise ValueError("Error return code cannot be 0.")
    if log is None:
        log = logging.getLogger(__name__)
    if output_path:
        if os.path.exists(output_path) and not overwrite:
            log.error("Output configuration file path already exists! (%s)",
                      output_path)
            sys.exit(error_rc)
        else:
            log.info("Outputting JSON configuration to: %s", output_path)
            with open(output_path, 'w') as f:
                json.dump(config_dict, f, indent=4, check_circular=True,
                          separators=(',', ': '), sort_keys=True)
            sys.exit(0)


class ProgressReporter (SmqtkObject):
    """
    Helper utility for reporting the state of a loop and the rate at which
    looping is occurring based on lapsed wall-time and a given reporting
    interval.

    Includes optional methods that are thread-safe.

    TODO: Add parameter for an optionally known total number of increments.
    """

    def __init__(self, log_func, interval):
        """
        Initialize this reporter.

        :param log_func: Logging function to use.
        :type log_func: (str, *args, **kwds) -> None

        :param interval: Time interval to perform reporting in seconds.  If no
            reporting during incrementation should occur, infinity should be
            passed.
        :type interval: float

        """
        self.log_func = log_func
        self.interval = float(interval)

        self.lock = threading.RLock()
        self.c_last = self.c = self.c_delta = 0
        self.t_last = self.t = self.t_delta = self.t_start = 0.0

        self.started = False

    def start(self):
        """ Start the timing state of this reporter.

        Repeated calls to this method resets the state of the reporting for
        multiple uses.

        This method is thread-safe.

        :returns: Self
        :rtype: ProgressReporter

        """
        with self.lock:
            self.started = True
            self.c_last = self.c = self.c_delta = 0
            self.t_last = self.t = self.t_start = time.time()
            self.t_delta = 0.0
        return self

    def increment_report(self):
        """
        Increment counter and time since last report, reporting if delta exceeds
        the set reporting interval period.
        """
        if not self.started:
            raise RuntimeError("Reporter needs to be started first.")
        self.c += 1
        self.t = time.time()
        self.t_delta = self.t - self.t_last
        # Only report if its been ``interval`` seconds since the last
        # report.
        if self.t_delta >= self.interval:
            self.c_delta = self.c - self.c_last
            self.report()
            self.t_last = self.t
            self.c_last = self.c

    def increment_report_threadsafe(self):
        """
        The same as ``increment_report`` but additionally acquires a lock on
        resources first for thread-safety.

        This version of the method is a little more costly due to the lock
        acquisition.
        """
        with self.lock:
            self.increment_report()

    def report(self):
        """
        Report the current state.

        Does nothing if no increments have occurred yet.
        """
        if not self.started:
            raise RuntimeError("Reporter needs to be started first.")
        # divide-by-zero safeguard
        if self.t_delta > 0 and (self.t - self.t_start) > 0:
            self.log_func("Loops per second %f (avg %f) "
                          "(%d current interval / %d total)"
                          % (self.c_delta / self.t_delta,
                             self.c / (self.t - self.t_start),
                             self.c_delta,
                             self.c))

    def report_threadsafe(self):
        """
        The same as ``report`` but additionally acquires a lock on
        resources first for thread-safety.

        This version of the method is a little more costly due to the lock
        acquisition.
        """
        with self.lock:
            self.report()


def report_progress(log, state, interval):
    """
    Loop progress reporting function that logs (when in debug) loops per
    second, loops in the last reporting period and total loops executed.

    The ``state`` given to this function must be a list of 7 integers, initially
    all set to 0. This function will update the fields of the state as its is
    called to control when reporting should happen and what to report.

    A report can be effectively force for a call by setting ``state[3] = 0``
    or ``interval`` to ``0``.

    :param log: Logger logging function to use to send reporting message to.
    :type log: (str, *args, **kwargs) -> None

    :param state: Reporting state. This should be initialized to a list of 6
        zeros (floats), and then should not be modified externally from this
        function.
    :type state: list[float]

    :param interval: Frequency in seconds that reporting messages should be
        made. This should be greater than 0.
    :type interval: float

    """
    warnings.warn("``report_progress`` is deprecated. Please use the"
                  "``ProgressReporter`` class instead.",
                  DeprecationWarning)
    # State format (c=count, t=time:
    #   [last_c, c, delta_c, last_t, t, delta_t, starting_t]
    #   [  0,    1,    2,       3,   4,    5,         6    ]

    # Starting time
    if not state[6]:
        state[3] = state[6] = time.time()

    state[1] += 1
    state[4] = time.time()
    state[5] = state[4] - state[3]
    if state[5] >= interval:
        state[2] = state[1] - state[0]
        # TODO: Could possibly to something with ncurses
        #       - to maintain a single line.
        log("Loops per second %f (avg %f) (%d this interval / %d total)"
            % (state[2] / state[5],
               state[1] / (state[4] - state[6]),
               state[2], state[1]))
        state[3] = state[4]
        state[0] = state[1]


def basic_cli_parser(description=None, configuration_group=True):
    """
    Generate an ``argparse.ArgumentParser`` with the given description and the
    basic options for verbosity and configuration/generation paths.

    The returned parser instance has an option for extra verbosity
    (-v/--verbose) and a group for configuration specification (-c/--config and
    configuration generation (-g/--generate-config) if enabled (true by
    default).

    :param description: Optional description string for the parser.
    :type description: str

    :param configuration_group: Whether or not to include the configuration
        group options.
    :type configuration_group: bool

    :return: Argument parser instance with basic options.
    :rtype: argparse.ArgumentParser

    """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument('-v', '--verbose',
                        default=False, action='store_true',
                        help='Output additional debug logging.')

    if configuration_group:
        g_config = parser.add_argument_group('Configuration')
        g_config.add_argument('-c', '--config',
                              metavar="PATH",
                              help='Path to the JSON configuration file.')
        g_config.add_argument('-g', '--generate-config',
                              metavar="PATH",
                              help='Optionally generate a default '
                                   'configuration file at the specified path. '
                                   'If a configuration file was provided, we '
                                   'update the default configuration with the '
                                   'contents of the given configuration.')

    return parser


def utility_main_helper(default_config, args, additional_logging_domains=(),
                        skip_logging_init=False, default_config_valid=False):
    """
    Helper function for utilities standardizing logging initialization, CLI
    parsing and configuration loading/generation.

    Specific utilities should use this in their main function. This
    encapsulates the following standard actions:

        - using ``argparse`` parser results to drive logging initialization
          (can be skipped if initialized externally)
        - handling loaded configuration merger onto the default
        - handling configuration generation based on given default and possibly
          specified input config.

    :param default_config: Function returning default configuration (JSON)
        dictionary for the utility. This should take no arguments.
    :type default_config: () -> dict

    :param args: Parsed arguments from argparse.ArgumentParser instance as
        returned from ``parser.parse_args()``.
    :type args: argparse.Namespace

    :param additional_logging_domains: We initialize logging on the base
        ``smqtk`` and ``__main__`` namespace. Any additional namespaces under
        which logging should be reported should be added here as an iterable.
    :type additional_logging_domains: collections.Iterable[str]

    :param skip_logging_init: Skip initialize logging in this function because
        it is done elsewhere externally.
    :type skip_logging_init: bool

    :param default_config_valid: Whether the default config returned from the
        generator is a valid config to continue execution with or not.
    :type default_config_valid: bool

    :return: Loaded configuration dictionary.
    :rtype: dict

    """
    # noinspection PyUnresolvedReferences
    config_filepath = args.config
    # noinspection PyUnresolvedReferences
    config_generate = args.generate_config
    # noinspection PyUnresolvedReferences
    verbose = args.verbose

    if not skip_logging_init:
        llevel = logging.INFO
        if verbose:
            llevel = logging.DEBUG
        initialize_logging(logging.getLogger('smqtk'), llevel)
        initialize_logging(logging.getLogger('__main__'), llevel)
        for d in additional_logging_domains:
            initialize_logging(logging.getLogger(d), llevel)

    config, config_loaded = load_config(config_filepath, default_config())
    output_config(config_generate, config, overwrite=True)

    if not (config_loaded or default_config_valid):
        raise RuntimeError("No configuration loaded (not trusting default).")

    return config
